"""
Elements
========

Defines an :class:`.Atom` class for each element.

"""

from __future__ import annotations

from typing import ClassVar, TypeVar

from stk._internal.atom import Atom

_T = TypeVar("_T", bound="AtomImpl")


class AtomImpl(Atom):
    """
    An implementation of the :class:`.Atom` interface.

    """

    _atomic_number: ClassVar[int]

    def __init_subclass__(cls: type[AtomImpl], **kwargs) -> None:
        cls._elements[cls._atomic_number] = cls

    def __init__(
        self,
        id: int,
        charge: int = 0,
    ) -> None:
        self._id = id
        self._charge = charge

    def get_id(self) -> int:
        return self._id

    def _with_id(self: _T, id: int) -> _T:
        """
        Modify the atom.

        """

        self._id = id
        return self

    def with_id(self, id: int) -> AtomImpl:
        return self.clone()._with_id(id)

    def get_atomic_number(self) -> int:
        return self._atomic_number

    def get_charge(self) -> int:
        return self._charge

    def clone(self) -> AtomImpl:
        return type(self)(self._id, self._charge)

    def __repr__(self) -> str:
        charge = f", charge={self._charge}" if self._charge != 0 else ""
        return f"{self.__class__.__name__}({self._id}{charge})"

    def __str__(self) -> str:
        return repr(self)


class H(AtomImpl):
    _atomic_number = 1

    def clone(self) -> H:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> H:
        return self.clone()._with_id(id)


class He(AtomImpl):
    _atomic_number = 2

    def clone(self) -> He:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> He:
        return self.clone()._with_id(id)


class Li(AtomImpl):
    _atomic_number = 3

    def clone(self) -> Li:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Li:
        return self.clone()._with_id(id)


class Be(AtomImpl):
    _atomic_number = 4

    def clone(self) -> Be:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Be:
        return self.clone()._with_id(id)


class B(AtomImpl):
    _atomic_number = 5

    def clone(self) -> B:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> B:
        return self.clone()._with_id(id)


class C(AtomImpl):
    _atomic_number = 6

    def clone(self) -> C:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> C:
        return self.clone()._with_id(id)


class N(AtomImpl):
    _atomic_number = 7

    def clone(self) -> N:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> N:
        return self.clone()._with_id(id)


# "O" is a valid elemental symbol.
class O(AtomImpl):  # noqa
    _atomic_number = 8

    def clone(self) -> O:  # noqa
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> O:  # noqa
        return self.clone()._with_id(id)


class F(AtomImpl):
    _atomic_number = 9

    def clone(self) -> F:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> F:
        return self.clone()._with_id(id)


class Ne(AtomImpl):
    _atomic_number = 10

    def clone(self) -> Ne:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Ne:
        return self.clone()._with_id(id)


class Na(AtomImpl):
    _atomic_number = 11

    def clone(self) -> Na:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Na:
        return self.clone()._with_id(id)


class Mg(AtomImpl):
    _atomic_number = 12

    def clone(self) -> Mg:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Mg:
        return self.clone()._with_id(id)


class Al(AtomImpl):
    _atomic_number = 13

    def clone(self) -> Al:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Al:
        return self.clone()._with_id(id)


class Si(AtomImpl):
    _atomic_number = 14

    def clone(self) -> Si:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Si:
        return self.clone()._with_id(id)


class P(AtomImpl):
    _atomic_number = 15

    def clone(self) -> P:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> P:
        return self.clone()._with_id(id)


class S(AtomImpl):
    _atomic_number = 16

    def clone(self) -> S:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> S:
        return self.clone()._with_id(id)


class Cl(AtomImpl):
    _atomic_number = 17

    def clone(self) -> Cl:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Cl:
        return self.clone()._with_id(id)


class Ar(AtomImpl):
    _atomic_number = 18

    def clone(self) -> Ar:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Ar:
        return self.clone()._with_id(id)


class K(AtomImpl):
    _atomic_number = 19

    def clone(self) -> K:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> K:
        return self.clone()._with_id(id)


class Ca(AtomImpl):
    _atomic_number = 20

    def clone(self) -> Ca:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Ca:
        return self.clone()._with_id(id)


class Sc(AtomImpl):
    _atomic_number = 21

    def clone(self) -> Sc:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Sc:
        return self.clone()._with_id(id)


class Ti(AtomImpl):
    _atomic_number = 22

    def clone(self) -> Ti:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Ti:
        return self.clone()._with_id(id)


class V(AtomImpl):
    _atomic_number = 23

    def clone(self) -> V:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> V:
        return self.clone()._with_id(id)


class Cr(AtomImpl):
    _atomic_number = 24

    def clone(self) -> Cr:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Cr:
        return self.clone()._with_id(id)


class Mn(AtomImpl):
    _atomic_number = 25

    def clone(self) -> Mn:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Mn:
        return self.clone()._with_id(id)


class Fe(AtomImpl):
    _atomic_number = 26

    def clone(self) -> Fe:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Fe:
        return self.clone()._with_id(id)


class Co(AtomImpl):
    _atomic_number = 27

    def clone(self) -> Co:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Co:
        return self.clone()._with_id(id)


class Ni(AtomImpl):
    _atomic_number = 28

    def clone(self) -> Ni:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Ni:
        return self.clone()._with_id(id)


class Cu(AtomImpl):
    _atomic_number = 29

    def clone(self) -> Cu:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Cu:
        return self.clone()._with_id(id)


class Zn(AtomImpl):
    _atomic_number = 30

    def clone(self) -> Zn:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Zn:
        return self.clone()._with_id(id)


class Ga(AtomImpl):
    _atomic_number = 31

    def clone(self) -> Ga:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Ga:
        return self.clone()._with_id(id)


class Ge(AtomImpl):
    _atomic_number = 32

    def clone(self) -> Ge:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Ge:
        return self.clone()._with_id(id)


class As(AtomImpl):
    _atomic_number = 33

    def clone(self) -> As:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> As:
        return self.clone()._with_id(id)


class Se(AtomImpl):
    _atomic_number = 34

    def clone(self) -> Se:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Se:
        return self.clone()._with_id(id)


class Br(AtomImpl):
    _atomic_number = 35

    def clone(self) -> Br:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Br:
        return self.clone()._with_id(id)


class Kr(AtomImpl):
    _atomic_number = 36

    def clone(self) -> Kr:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Kr:
        return self.clone()._with_id(id)


class Rb(AtomImpl):
    _atomic_number = 37

    def clone(self) -> Rb:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Rb:
        return self.clone()._with_id(id)


class Sr(AtomImpl):
    _atomic_number = 38

    def clone(self) -> Sr:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Sr:
        return self.clone()._with_id(id)


class Y(AtomImpl):
    _atomic_number = 39

    def clone(self) -> Y:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Y:
        return self.clone()._with_id(id)


class Zr(AtomImpl):
    _atomic_number = 40

    def clone(self) -> Zr:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Zr:
        return self.clone()._with_id(id)


class Nb(AtomImpl):
    _atomic_number = 41

    def clone(self) -> Nb:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Nb:
        return self.clone()._with_id(id)


class Mo(AtomImpl):
    _atomic_number = 42

    def clone(self) -> Mo:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Mo:
        return self.clone()._with_id(id)


class Tc(AtomImpl):
    _atomic_number = 43

    def clone(self) -> Tc:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Tc:
        return self.clone()._with_id(id)


class Ru(AtomImpl):
    _atomic_number = 44

    def clone(self) -> Ru:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Ru:
        return self.clone()._with_id(id)


class Rh(AtomImpl):
    _atomic_number = 45

    def clone(self) -> Rh:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Rh:
        return self.clone()._with_id(id)


class Pd(AtomImpl):
    _atomic_number = 46

    def clone(self) -> Pd:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Pd:
        return self.clone()._with_id(id)


class Ag(AtomImpl):
    _atomic_number = 47

    def clone(self) -> Ag:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Ag:
        return self.clone()._with_id(id)


class Cd(AtomImpl):
    _atomic_number = 48

    def clone(self) -> Cd:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Cd:
        return self.clone()._with_id(id)


class In(AtomImpl):
    _atomic_number = 49

    def clone(self) -> In:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> In:
        return self.clone()._with_id(id)


class Sn(AtomImpl):
    _atomic_number = 50

    def clone(self) -> Sn:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Sn:
        return self.clone()._with_id(id)


class Sb(AtomImpl):
    _atomic_number = 51

    def clone(self) -> Sb:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Sb:
        return self.clone()._with_id(id)


class Te(AtomImpl):
    _atomic_number = 52

    def clone(self) -> Te:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Te:
        return self.clone()._with_id(id)


# "I" is a valid elemental symbol.
class I(AtomImpl):  # noqa
    _atomic_number = 53

    def clone(self) -> I:  # noqa
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> I:  # noqa
        return self.clone()._with_id(id)


class Xe(AtomImpl):
    _atomic_number = 54

    def clone(self) -> Xe:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Xe:
        return self.clone()._with_id(id)


class Cs(AtomImpl):
    _atomic_number = 55

    def clone(self) -> Cs:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Cs:
        return self.clone()._with_id(id)


class Ba(AtomImpl):
    _atomic_number = 56

    def clone(self) -> Ba:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Ba:
        return self.clone()._with_id(id)


class La(AtomImpl):
    _atomic_number = 57

    def clone(self) -> La:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> La:
        return self.clone()._with_id(id)


class Ce(AtomImpl):
    _atomic_number = 58

    def clone(self) -> Ce:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Ce:
        return self.clone()._with_id(id)


class Pr(AtomImpl):
    _atomic_number = 59

    def clone(self) -> Pr:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Pr:
        return self.clone()._with_id(id)


class Nd(AtomImpl):
    _atomic_number = 60

    def clone(self) -> Nd:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Nd:
        return self.clone()._with_id(id)


class Pm(AtomImpl):
    _atomic_number = 61

    def clone(self) -> Pm:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Pm:
        return self.clone()._with_id(id)


class Sm(AtomImpl):
    _atomic_number = 62

    def clone(self) -> Sm:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Sm:
        return self.clone()._with_id(id)


class Eu(AtomImpl):
    _atomic_number = 63

    def clone(self) -> Eu:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Eu:
        return self.clone()._with_id(id)


class Gd(AtomImpl):
    _atomic_number = 64

    def clone(self) -> Gd:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Gd:
        return self.clone()._with_id(id)


class Tb(AtomImpl):
    _atomic_number = 65

    def clone(self) -> Tb:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Tb:
        return self.clone()._with_id(id)


class Dy(AtomImpl):
    _atomic_number = 66

    def clone(self) -> Dy:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Dy:
        return self.clone()._with_id(id)


class Ho(AtomImpl):
    _atomic_number = 67

    def clone(self) -> Ho:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Ho:
        return self.clone()._with_id(id)


class Er(AtomImpl):
    _atomic_number = 68

    def clone(self) -> Er:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Er:
        return self.clone()._with_id(id)


class Tm(AtomImpl):
    _atomic_number = 69

    def clone(self) -> Tm:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Tm:
        return self.clone()._with_id(id)


class Yb(AtomImpl):
    _atomic_number = 70

    def clone(self) -> Yb:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Yb:
        return self.clone()._with_id(id)


class Lu(AtomImpl):
    _atomic_number = 71

    def clone(self) -> Lu:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Lu:
        return self.clone()._with_id(id)


class Hf(AtomImpl):
    _atomic_number = 72

    def clone(self) -> Hf:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Hf:
        return self.clone()._with_id(id)


class Ta(AtomImpl):
    _atomic_number = 73

    def clone(self) -> Ta:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Ta:
        return self.clone()._with_id(id)


class W(AtomImpl):
    _atomic_number = 74

    def clone(self) -> W:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> W:
        return self.clone()._with_id(id)


class Re(AtomImpl):
    _atomic_number = 75

    def clone(self) -> Re:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Re:
        return self.clone()._with_id(id)


class Os(AtomImpl):
    _atomic_number = 76

    def clone(self) -> Os:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Os:
        return self.clone()._with_id(id)


class Ir(AtomImpl):
    _atomic_number = 77

    def clone(self) -> Ir:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Ir:
        return self.clone()._with_id(id)


class Pt(AtomImpl):
    _atomic_number = 78

    def clone(self) -> Pt:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Pt:
        return self.clone()._with_id(id)


class Au(AtomImpl):
    _atomic_number = 79

    def clone(self) -> Au:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Au:
        return self.clone()._with_id(id)


class Hg(AtomImpl):
    _atomic_number = 80

    def clone(self) -> Hg:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Hg:
        return self.clone()._with_id(id)


class Tl(AtomImpl):
    _atomic_number = 81

    def clone(self) -> Tl:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Tl:
        return self.clone()._with_id(id)


class Pb(AtomImpl):
    _atomic_number = 82

    def clone(self) -> Pb:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Pb:
        return self.clone()._with_id(id)


class Bi(AtomImpl):
    _atomic_number = 83

    def clone(self) -> Bi:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Bi:
        return self.clone()._with_id(id)


class Po(AtomImpl):
    _atomic_number = 84

    def clone(self) -> Po:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Po:
        return self.clone()._with_id(id)


class At(AtomImpl):
    _atomic_number = 85

    def clone(self) -> At:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> At:
        return self.clone()._with_id(id)


class Rn(AtomImpl):
    _atomic_number = 86

    def clone(self) -> Rn:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Rn:
        return self.clone()._with_id(id)


class Fr(AtomImpl):
    _atomic_number = 87

    def clone(self) -> Fr:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Fr:
        return self.clone()._with_id(id)


class Ra(AtomImpl):
    _atomic_number = 88

    def clone(self) -> Ra:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Ra:
        return self.clone()._with_id(id)


class Ac(AtomImpl):
    _atomic_number = 89

    def clone(self) -> Ac:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Ac:
        return self.clone()._with_id(id)


class Th(AtomImpl):
    _atomic_number = 90

    def clone(self) -> Th:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Th:
        return self.clone()._with_id(id)


class Pa(AtomImpl):
    _atomic_number = 91

    def clone(self) -> Pa:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Pa:
        return self.clone()._with_id(id)


class U(AtomImpl):
    _atomic_number = 92

    def clone(self) -> U:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> U:
        return self.clone()._with_id(id)


class Np(AtomImpl):
    _atomic_number = 93

    def clone(self) -> Np:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Np:
        return self.clone()._with_id(id)


class Pu(AtomImpl):
    _atomic_number = 94

    def clone(self) -> Pu:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Pu:
        return self.clone()._with_id(id)


class Am(AtomImpl):
    _atomic_number = 95

    def clone(self) -> Am:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Am:
        return self.clone()._with_id(id)


class Cm(AtomImpl):
    _atomic_number = 96

    def clone(self) -> Cm:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Cm:
        return self.clone()._with_id(id)


class Bk(AtomImpl):
    _atomic_number = 97

    def clone(self) -> Bk:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Bk:
        return self.clone()._with_id(id)


class Cf(AtomImpl):
    _atomic_number = 98

    def clone(self) -> Cf:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Cf:
        return self.clone()._with_id(id)


class Es(AtomImpl):
    _atomic_number = 99

    def clone(self) -> Es:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Es:
        return self.clone()._with_id(id)


class Fm(AtomImpl):
    _atomic_number = 100

    def clone(self) -> Fm:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Fm:
        return self.clone()._with_id(id)


class Md(AtomImpl):
    _atomic_number = 101

    def clone(self) -> Md:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Md:
        return self.clone()._with_id(id)


class No(AtomImpl):
    _atomic_number = 102

    def clone(self) -> No:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> No:
        return self.clone()._with_id(id)


class Lr(AtomImpl):
    _atomic_number = 103

    def clone(self) -> Lr:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Lr:
        return self.clone()._with_id(id)


class Rf(AtomImpl):
    _atomic_number = 104

    def clone(self) -> Rf:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Rf:
        return self.clone()._with_id(id)


class Db(AtomImpl):
    _atomic_number = 105

    def clone(self) -> Db:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Db:
        return self.clone()._with_id(id)


class Sg(AtomImpl):
    _atomic_number = 106

    def clone(self) -> Sg:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Sg:
        return self.clone()._with_id(id)


class Bh(AtomImpl):
    _atomic_number = 107

    def clone(self) -> Bh:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Bh:
        return self.clone()._with_id(id)


class Hs(AtomImpl):
    _atomic_number = 108

    def clone(self) -> Hs:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Hs:
        return self.clone()._with_id(id)


class Mt(AtomImpl):
    _atomic_number = 109

    def clone(self) -> Mt:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Mt:
        return self.clone()._with_id(id)


class Ds(AtomImpl):
    _atomic_number = 110

    def clone(self) -> Ds:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Ds:
        return self.clone()._with_id(id)


class Rg(AtomImpl):
    _atomic_number = 111

    def clone(self) -> Rg:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Rg:
        return self.clone()._with_id(id)


class Cn(AtomImpl):
    _atomic_number = 112

    def clone(self) -> Cn:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Cn:
        return self.clone()._with_id(id)


class Nh(AtomImpl):
    _atomic_number = 113

    def clone(self) -> Nh:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Nh:
        return self.clone()._with_id(id)


class Fl(AtomImpl):
    _atomic_number = 114

    def clone(self) -> Fl:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Fl:
        return self.clone()._with_id(id)


class Mc(AtomImpl):
    _atomic_number = 115

    def clone(self) -> Mc:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Mc:
        return self.clone()._with_id(id)


class Lv(AtomImpl):
    _atomic_number = 116

    def clone(self) -> Lv:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Lv:
        return self.clone()._with_id(id)


class Ts(AtomImpl):
    _atomic_number = 117

    def clone(self) -> Ts:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Ts:
        return self.clone()._with_id(id)


class Og(AtomImpl):
    _atomic_number = 118

    def clone(self) -> Og:
        return type(self)(self._id, self._charge)

    def with_id(self, id: int) -> Og:
        return self.clone()._with_id(id)



from typing import Any

import pytest
import stk

from ..case_data import CaseData


def get_topology_graph(num_repeating_units: int) -> stk.TopologyGraph:
    return stk.polymer.Linear(
        building_blocks=[
            stk.BuildingBlock("BrCCBr", stk.BromoFactory()),
        ],
        repeating_unit="A",
        num_repeating_units=num_repeating_units,
    )


@pytest.fixture(scope="session")
def remove_batches_population_1() -> dict[stk.MoleculeRecord[Any], float]:
    return {
        stk.MoleculeRecord(
            topology_graph=get_topology_graph(2),
        ): 10,
        stk.MoleculeRecord(
            topology_graph=get_topology_graph(3),
        ): 9,
        stk.MoleculeRecord(
            topology_graph=get_topology_graph(4),
        ): 2,
        stk.MoleculeRecord(
            topology_graph=get_topology_graph(5),
        ): 1,
        stk.MoleculeRecord(
            topology_graph=get_topology_graph(6),
        ): 1,
    }


@pytest.fixture(
    scope="session",
    params=(
        lambda population: CaseData.new(
            selector=stk.RemoveBatches(
                remover=stk.Worst(4),
                selector=stk.Best(),
            ),
            population=population,
            selected=[[0]],
        ),
    ),
)
def remove_batches(
    request: pytest.FixtureRequest,
    remove_batches_population_1: dict[stk.MoleculeRecord[Any], float],
) -> CaseData:
    return request.param(remove_batches_population_1)



from __future__ import annotations

import pytest
import stk

from .utilities import is_clone


@pytest.mark.skip
def test_with_vertices(construction_state, vertices):
    clone = construction_state.clone()
    _test_with_vertices(construction_state, vertices)
    # Test immutability.
    is_clone(construction_state, clone)


def _test_with_vertices(construction_state, vertices):
    clone = construction_state.with_vertices(vertices)
    assert clone.get_num_vertices() == len(vertices)
    for vertex_id in range(len(vertices)):
        assert clone.get_vertex(vertex_id) is vertices[vertex_id]


@pytest.fixture(
    scope="session",
    params=(
        (
            lambda: stk.Vertex(0, [0, 0, 0]),
            lambda: stk.Vertex(1, [10, 0, 0]),
            lambda: stk.Vertex(2, [20, 0, 0]),
        ),
    ),
)
def vertices(request) -> tuple[stk.Vertex, ...]:
    return request.param()



"""
Topology Graph Implementation Utilities
=======================================

"""


import typing

import numpy as np

if typing.TYPE_CHECKING:
    from stk._internal.construction_state.construction_state import (
        ConstructionState,
    )


class _TopologyGraphImplementation(typing.Protocol):
    def _place_building_blocks(
        self, state: "ConstructionState"
    ) -> "ConstructionState":
        pass

    def get_num_stages(self) -> int:
        pass


class _PlacementResult(typing.NamedTuple):
    """
    The result of a building block placement.

    Attributes:

        position_matrix:
            The position matrix of the building block after it has been
            placed on the vertex.

        functional_group_edges:
            Maps the id of a functional group to the id of the edge it
            is assigned to.

    """

    position_matrix: np.ndarray
    functional_group_edges: dict[int, int]


class _Placement:
    """
    Represents placement of a building block on a vertex.

    It represents a computation which carries out the placement
    of the building block on the vertex and the mapping of its
    functional group to edges.

    """

    def __init__(self, vertex, edges, building_block):
        """
        Initialize a :class:`._Placement`.

        Parameters
        ----------
        vertex : :class:`.Vertex`
            The vertex which does the placement.

        edges : :class:`tuple` of :class:`.Edge`
            The edges connected to `vertex`.

        building_block : :class:`.BuildingBlock`
            The building block to be placed on `vertex`.

        """

        self._vertex = vertex
        self._edges = edges
        self._building_block = building_block

    def get_result(self):
        """
        Get the result of the placement.

        Returns
        -------
        :class:`_PlacementResult`
            The result of the placement.

        """

        position_matrix = self._vertex.place_building_block(
            building_block=self._building_block,
            edges=self._edges,
        )
        position_matrix.setflags(write=False)
        building_block = self._building_block.with_position_matrix(
            position_matrix=position_matrix,
        )
        functional_group_edges = self._vertex.map_functional_groups_to_edges(
            building_block=building_block,
            edges=self._edges,
        )
        return _PlacementResult(
            position_matrix=position_matrix,
            functional_group_edges=functional_group_edges,
        )



import pytest
import stk

from ....case_data import CaseData


def _get_zinc_atom() -> stk.BuildingBlock:
    return stk.BuildingBlock(
        smiles="[Zn+2]",
        functional_groups=(
            stk.SingleAtom(stk.Zn(0, charge=2)) for i in range(4)
        ),
        position_matrix=([0, 0, 0],),
    )


def _get_quad_1() -> stk.BuildingBlock:
    return stk.BuildingBlock(
        smiles=(
            "Brc1ccc(C2=C3C=CC(=C(c4ccc(Br)cc4)C4=NC(=C(c5ccc(Br)cc5)"
            "C5=CC=C([N]5)C(c5ccc(Br)cc5)=C5C=CC2=N5)C=C4)[N]3)cc1"
        ),
        functional_groups=[
            stk.SmartsFunctionalGroupFactory(
                smarts="[#6]~[#7X2]~[#6]",
                bonders=(1,),
                deleters=(),
            ),
        ],
    )


@pytest.fixture(
    scope="session",
    params=(
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                stk.metal_complex.Porphyrin(
                    metals={_get_zinc_atom(): 0},
                    ligands={_get_quad_1(): 0},
                    reaction_factory=stk.DativeReactionFactory(
                        stk.GenericReactionFactory(
                            bond_orders={
                                frozenset(
                                    {
                                        stk.GenericFunctionalGroup,
                                        stk.SingleAtom,
                                    }
                                ): 9,
                            },
                        ),
                    ),
                ),
            ),
            smiles=(
                "[H]C1=C([H])C(C2=C3C([H])=C([H])C4=N3->[Zn+2]35<-N"
                "6=C(C([H])=C([H])C6=C(C6=C([H])C([H])=C(Br)C([H])=C6"
                "[H])C6=C([H])C([H])=C2N->36)C(C2=C([H])C([H])=C(Br)C"
                "([H])=C2[H])=C2C([H])=C([H])C(=C4C3=C([H])C([H])=C(Br"
                ")C([H])=C3[H])N->52)=C([H])C([H])=C1Br"
            ),
            name=name,
        ),
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                stk.metal_complex.Porphyrin(
                    metals=_get_zinc_atom(),
                    ligands=_get_quad_1(),
                    reaction_factory=stk.DativeReactionFactory(
                        stk.GenericReactionFactory(
                            bond_orders={
                                frozenset(
                                    {
                                        stk.GenericFunctionalGroup,
                                        stk.SingleAtom,
                                    }
                                ): 9,
                            },
                        ),
                    ),
                ),
            ),
            smiles=(
                "[H]C1=C([H])C(C2=C3C([H])=C([H])C4=N3->[Zn+2]35<-N"
                "6=C(C([H])=C([H])C6=C(C6=C([H])C([H])=C(Br)C([H])=C6"
                "[H])C6=C([H])C([H])=C2N->36)C(C2=C([H])C([H])=C(Br)C"
                "([H])=C2[H])=C2C([H])=C([H])C(=C4C3=C([H])C([H])=C(Br"
                ")C([H])=C3[H])N->52)=C([H])C([H])=C1Br"
            ),
            name=name,
        ),
    ),
)
def metal_complex_porphyrin(request) -> CaseData:
    return request.param(
        f"{request.fixturename}{request.param_index}",
    )



from tests.utilities import is_equivalent


def test_cross(case_data):
    """
    Test :meth:`cross`.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        A test case. Holds the crosser to test and the correct
        crossover records.

    Returns
    -------
    None : :class:`NoneType`

    """

    _test_cross(
        crosser=case_data.crosser,
        records=case_data.records,
        crossover_records=case_data.crossover_records,
    )


def _test_cross(crosser, records, crossover_records):
    """
    Test :meth:`cross`.

    Parameters
    ----------
    crosser : :class:`.MoleculeCrosser`
        The crosser to test.

    records : :class:`tuple` of :class:`.MoleculeRecord`
        The molecules to cross.

    crossover_records : :class:`tuple` of :class:`.CrossoverRecord`
        The correct offspring.

    Returns
    -------
    None : :class:`NoneType`

    """

    for record1, record2 in zip(
        crosser.cross(records),
        crossover_records,
    ):
        assert record1.get_crosser_name() == record2.get_crosser_name()
        is_equivalent(
            record1.get_molecule_record().get_molecule(),
            record2.get_molecule_record().get_molecule(),
        )



from dataclasses import dataclass
from typing import Callable

import pymongo
import pytest
import stk

from ..case_data import CaseData


@dataclass(frozen=True)
class CaseDataData:
    """
    Data used to create a :class:`.CaseData` instance.

    Attributes
    ----------
    get_database : :class:`callable`
        Creates the database to test. Takes a
        :class:`pymongo.MongoClient` as input and returns a
        :class:`.ValueMongoDb` instance.

    molecule : :class:`.Molecule`
        The molecule to test.

    value : :class:`object`
        The value to put into the database.

    """

    get_database: Callable[[pymongo.MongoClient], stk.ValueMongoDb]
    molecule: stk.Molecule
    value: object


@pytest.fixture(
    params=(
        lambda: CaseDataData(
            get_database=lambda mongo_client: stk.ValueMongoDb(
                mongo_client=mongo_client,
                collection="values",
                database="_stk_test_database_for_testing",
                put_lru_cache_size=0,
                get_lru_cache_size=0,
            ),
            molecule=stk.BuildingBlock("BrCCBr"),
            value=12,
        ),
        lambda: CaseDataData(
            get_database=lambda mongo_client: stk.ValueMongoDb(
                mongo_client=mongo_client,
                collection="values",
                database="_stk_test_database_for_testing",
                put_lru_cache_size=128,
                get_lru_cache_size=128,
            ),
            molecule=stk.BuildingBlock("BrCCBr"),
            value=12,
        ),
    ),
)
def mongo_db(
    request,
    mongo_client: pymongo.MongoClient,
) -> CaseData:
    data = request.param()
    return CaseData(
        database=data.get_database(mongo_client),
        molecule=data.molecule,
        value=data.value,
    )



"""
Atom
====

"""

from __future__ import annotations


class Atom:
    """
    An abstract base class for atoms.

    A subclass is made for each element. The name of each subclass is
    the periodic table symbol of that element.

    Atoms of a particular element can be made with this
    class or with the subclass representing that element.

    Examples:

        *Initialization.*

        Initialization of an :class:`.Atom` can happen in one of two
        ways. The atom can be initialized through the :class:`.Atom`
        class or through the class representing the element.

        .. testcode:: initialization

            import stk

            # h0 is an instance of the H class.
            h0 = stk.Atom(id=0, atomic_number=1)

            # h1 is also an instance of the H class.
            h1 = stk.H(id=1)

        When the class corresponding to the element is used directly,
        the ``atomic_number`` is not provided. Here are a few more
        examples.

        .. testcode:: initialization

            # Both he0 and he1 are instances of the He class.
            he0 = stk.Atom(id=2, atomic_number=2)
            he1 = stk.He(id=3)

            # Both c0 and c1 are instances of the
            # C class.
            c0 = stk.Atom(id=4, atomic_number=6)
            c1 = stk.C(id=5)

    """

    # Maps each atomic number (int) to the relevant Atom subclass.
    _elements: dict[int, type[Atom]] = {}

    def __init__(
        self,
        id: int,
        atomic_number: int,
        charge: int = 0,
    ) -> None:
        """
        Initialize an :class:`Atom`.

        Parameters:

            id: The id of the atom.

            atomic_number: The atomic number.

            charge: The formal charge.

        """

        self._elements[atomic_number].__init__(self, id, charge)
        self.__class__ = self._elements[atomic_number]

    def get_id(self) -> int:
        """
        Get the id of the atom.

        Returns:
            The id.

        """

        raise NotImplementedError()

    def with_id(self, id: int) -> Atom:
        """
        Get a clone but with a different id.

        Parameters:
            id: The id of the clone.

        Returns:
            A clone with a new id.

        """

        raise NotImplementedError()

    def get_atomic_number(self) -> int:
        """
        Get the atomic number of the atom.

        Returns:
            The atomic number.

        """

        raise NotImplementedError()

    def get_charge(self) -> int:
        """
        Get the charge of the atom.

        Returns:
            The charge.

        """

        raise NotImplementedError()

    def clone(self) -> Atom:
        """
        Return a clone.

        Returns:
            The clone.

        """

        raise NotImplementedError()



import numpy as np

from ...utilities import is_clone


def test_with_displacement(molecule, displacement):
    """
    Test :meth:`.Molecule.with_displacement`.

    Parameters
    ----------
    molecule : :class:`.Molecule`
        The molecule to test.

    displacement : :class:`numpy.ndarray`
        The displacement the clone should have.

    Returns
    -------
    None : :class:`NoneType`

    """

    new = molecule.with_displacement(displacement)
    is_clone(new, molecule)
    assert np.allclose(
        a=molecule.get_position_matrix() + displacement,
        b=new.get_position_matrix(),
        atol=1e-32,
    )



"""
XYZ Updating Utilities
======================

"""

import numpy as np

from stk._internal.utilities.utilities import periodic_table


def _with_structure_from_xyz(self, path):
    """
    Return a clone, with its structure taken from an ``.xyz`` file.

    Parameters
    ----------
    path : :class:`str`
        The full path of the ``.mol`` file from which the structure
        should be updated.

    Returns
    -------
    :class:`.Molecule`
        A clone with atomic positions found in `path`.

    Raises
    ------
    :class:`RuntimeError`
        If the number of atoms in the file does not match the
        number of atoms in the molecule or if atom elements in the
        file do not agree with the atom elements in the molecule.

    """

    with open(path, "r") as f:
        atom_count, _, *content = f.readlines()

    # Check the atom count is correct.
    num_atoms = len(self._atoms)
    if int(atom_count) != num_atoms:
        raise RuntimeError(
            f"The number of atoms in the xyz file, {atom_count}, "
            "does not match the number of atoms in the "
            f"molecule, {num_atoms}."
        )

    # Save all the coords in the file.
    new_coords = []
    for i, line in enumerate(content):
        element, *coords = line.split()
        # Handle XYZ files with capitilisation of element symbols.
        element = element.title()
        if element.isnumeric():
            element = periodic_table[int(element)]

        if element != self._atoms[i].__class__.__name__:
            raise RuntimeError(f"Atom {i} element does not match file.")

        new_coords.append([float(i) for i in coords])

    # Check that the correct number of atom
    # lines was present in the file.
    if i + 1 != num_atoms:
        raise RuntimeError(
            f"The number of atom lines in the xyz file, {i+1}, "
            "does not match the number of atoms in the "
            f"molecule, {num_atoms}."
        )

    # Update the structure.
    new_coords = np.array(new_coords)
    return self._with_position_matrix(new_coords)



class CaseData:
    """
    A test case.

    Attributes
    ----------
    molecule : :class:`.Molecule`
        The molecule to test.

    canonical_atom_ids : :class:`dict`
        The correct mapping of atom ids in :attr:`.molecule` to their
        canonical atom ids.

    """

    def __init__(self, molecule, canonical_atom_ids):
        """
        Initialize a :class:`.CaseData` instance.

        Parameters
        ----------
        molecule : :class:`.Molecule`
            The molecule to test.

        canonical_atom_ids : :class:`dict`
            The correct mapping of atom ids in :attr:`.molecule` to
            their canonical atom ids.

        """

        self.molecule = molecule
        self.canonical_atom_ids = canonical_atom_ids



"""
JSON Deserialization Utilities
==============================

"""

from stk._internal.atom import Atom
from stk._internal.atom_info import AtomInfo
from stk._internal.bond import Bond
from stk._internal.bond_info import BondInfo


def to_atom(id, json):
    return Atom(id, json[0], json[1])


def to_bond(atoms, json):
    return Bond(
        atom1=atoms[json[0]],
        atom2=atoms[json[1]],
        order=json[2],
        periodicity=tuple(json[3]),
    )


def to_atom_info(building_blocks, atom, json):
    if json[0] is None:
        return AtomInfo(
            atom=atom,
            building_block_atom=None,
            building_block=None,
            building_block_id=None,
        )

    building_block = building_blocks[json[0]]
    (building_block_atom,) = building_block.get_atoms(json[2])

    return AtomInfo(
        atom=atom,
        building_block_atom=building_block_atom,
        building_block=building_block,
        building_block_id=json[1],
    )


def to_bond_info(building_blocks, bond, json):
    return BondInfo(
        bond=bond,
        building_block=(
            building_blocks[json[0]] if json[0] is not None else None
        ),
        building_block_id=json[1],
    )


AtomicNumber = int
AtomicCharge = int
AtomJson = tuple[
    AtomicNumber,
    AtomicCharge,
]
AtomId = int
BondOrder = int
BondPeriodicity = tuple[int, int, int]
BondJson = tuple[
    AtomId,
    AtomId,
    BondOrder,
    BondPeriodicity,
]


def atom_to_json(
    atom: Atom,
) -> AtomJson:
    """
    Return a JSON representation of `atom`.

    Parameters:

        atom:
            The atom to serialize.

    Returns:

        A JSON representation of `atom`.

    """

    return (
        atom.get_atomic_number(),
        atom.get_charge(),
    )


def bond_to_json(
    bond: Bond,
) -> BondJson:
    """
    Return a JSON representation of `bond`.

    Parameters:

        bond:
            The bond to serialize.

    Returns:

        A JSON representation of `bond`.

    """

    return (
        bond.get_atom1().get_id(),
        bond.get_atom2().get_id(),
        bond.get_order(),
        bond.get_periodicity(),
    )



"""
Generic Reaction Factory
========================

"""

from stk._internal.functional_groups.aldehyde import Aldehyde
from stk._internal.functional_groups.alkene import Alkene
from stk._internal.functional_groups.alkyne import Alkyne
from stk._internal.functional_groups.amide import Amide
from stk._internal.functional_groups.primary_amino import PrimaryAmino
from stk._internal.reactions.one_one_reaction import OneOneReaction
from stk._internal.reactions.one_two_reaction import OneTwoReaction
from stk._internal.reactions.two_two_reaction import TwoTwoReaction

from .reaction_factory import ReactionFactory

# Impose the same interface on all reaction initializers.


def _one_one_reaction(
    construction_state,
    functional_group1,
    functional_group2,
    bond_order,
    periodicity,
):
    return OneOneReaction(
        functional_group1=functional_group1,
        functional_group2=functional_group2,
        bond_order=bond_order,
        periodicity=periodicity,
    )


def _one_two_reaction(
    construction_state,
    functional_group1,
    functional_group2,
    bond_order,
    periodicity,
):
    return OneTwoReaction(
        functional_group1=functional_group1,
        functional_group2=functional_group2,
        bond_order=bond_order,
        periodicity=periodicity,
    )


def _two_two_reaction(
    construction_state,
    functional_group1,
    functional_group2,
    bond_order,
    periodicity,
):
    return TwoTwoReaction(
        construction_state=construction_state,
        functional_group1=functional_group1,
        functional_group2=functional_group2,
        bond_order=bond_order,
        periodicity=periodicity,
    )


def _get_reaction_key(functional_groups):
    """
    Return a key for :data:`._reactions`

    Parameters
    ----------
    functional_groups : :class:`iterable`
        An :class:`iterable` of :class:`.GenericFunctionalGroup`.
        The correct reaction must be selected for these functional
        groups.

    Returns
    -------
    :class:`.frozenset`
        A key for :data:`_reactions`, which maps to the correct
        reaction.

    """

    return frozenset(
        functional_group.get_num_bonders()
        for functional_group in functional_groups
    )


_reactions = {
    frozenset({1}): _one_one_reaction,
    frozenset({1, 2}): _one_two_reaction,
    frozenset({2}): _two_two_reaction,
}


class GenericReactionFactory(ReactionFactory):
    """
    Create reactions for :class:`.GenericFunctionalGroup` instances.

    This reaction factory assumes that the functional groups, which
    belong to the :class:`.EdgeGroup` passed to it, are
    :class:`.GenericFunctionalGroup` instances. It returns a
    :class:`.Reaction` suitable for two such instances.

    """

    def __init__(self, bond_orders=None):
        """
        Initialize a :class:`.GenericReactionFactory`.

        Parameters
        ----------
        bond_orders : :class:`dict`, optional
            Maps a :class:`frozenset` of
            :class:`.GenericFunctionalGroup` subclasses to the
            bond orders for their respective reactions, if a pair
            of functional groups is missing, a default bond order of
            1 will be used for their reactions. If `bond_orders` is
            ``None``, the following :class:`dict` will be used

            .. testcode:: init

                import stk

                bond_orders = {
                    frozenset({stk.PrimaryAmino, stk.Aldehyde}): 2,
                    frozenset({stk.PrimaryAmino, stk.Aldehyde}): 2,
                    frozenset({stk.Amide, stk.PrimaryAmino}): 2,
                    frozenset({stk.Alkene}): 2,
                    frozenset({stk.Alkyne}): 2,
                }

            This means that if you want to get a reaction for
            an amine and an aldehyde functional group, the reaction
            will create bonds with a bond order of 2.

        """

        # Used for __repr__.
        self._default_bond_orders = bond_orders is None

        if bond_orders is None:
            bond_orders = {
                frozenset({PrimaryAmino, Aldehyde}): 2,
                frozenset({Amide, Aldehyde}): 2,
                frozenset({Amide, PrimaryAmino}): 2,
                frozenset({Alkene}): 2,
                frozenset({Alkyne}): 2,
            }

        self._bond_orders = bond_orders

    def get_reaction(self, construction_state, edge_group):
        functional_groups = tuple(
            construction_state.get_edge_group_functional_groups(
                edge_group=edge_group,
            )
        )
        functional_group1, functional_group2 = functional_groups
        edge = construction_state.get_edge(
            edge_id=next(edge_group.get_edge_ids()),
        )
        return _reactions[_get_reaction_key(functional_groups)](
            construction_state=construction_state,
            functional_group1=functional_group1,
            functional_group2=functional_group2,
            bond_order=self._bond_orders.get(
                self._get_bond_order_key(functional_groups),
                1,
            ),
            periodicity=edge.get_periodicity(),
        )

    def _get_bond_order_key(self, functional_groups):
        """
        Return a key for the :attr:`_bond_orders`.

        Parameters
        ----------
        functional_groups : :class:`iterable`
            A :class:`iterable` of :class:`.GenericFunctionalGroup`
            instances. You want to to get the bond order for a reaction
            involving these instances.

        Returns
        -------
        :class:`frozenset`
            A key for :attr:`_bond_orders`, which maps to the correct
            bond order.

        """

        return frozenset(map(type, functional_groups))

    def __str__(self):
        return repr(self)

    def __repr__(self):
        bond_orders = (
            "" if self._default_bond_orders else f"{self._bond_orders}"
        )
        return f"{self.__class__.__name__}({bond_orders})"



import pytest
import stk

from .case_data import CaseData


@pytest.fixture(
    params=(
        lambda: CaseData(
            molecule=stk.BuildingBlock("NCCN"),
            num_atoms=12,
        ),
    ),
)
def case_data(request) -> CaseData:
    return request.param()



import pytest
import stk

from .case_data import CaseData


@pytest.fixture(
    params=(
        lambda: CaseData(
            molecule=stk.BuildingBlock("NCCN"),
            atoms=(
                stk.N(0),
                stk.C(1),
                stk.C(2),
                stk.N(3),
                stk.H(4),
                stk.H(5),
                stk.H(6),
                stk.H(7),
                stk.H(8),
                stk.H(9),
                stk.H(10),
                stk.H(11),
            ),
        ),
    ),
)
def case_data(request) -> CaseData:
    """
    A :class:`.CaseData` instance.

    """

    return request.param()



import pytest
import stk

from ..case_data import GenericCaseData


@pytest.fixture
def alcohol(get_atom_ids):
    a, b, c = get_atom_ids(3)
    return _alcohol(stk.O(a), stk.H(b), stk.C(c))


def _alcohol(oxygen, hydrogen, atom):
    bonders = (oxygen,)
    deleters = (hydrogen,)
    return GenericCaseData(
        functional_group=stk.Alcohol(
            oxygen=oxygen,
            hydrogen=hydrogen,
            atom=atom,
            bonders=bonders,
            deleters=deleters,
        ),
        atoms=(oxygen, hydrogen, atom),
        bonders=bonders,
        deleters=deleters,
    )



"""
Edge Group
==========

"""


class EdgeGroup:
    """
    Represents a group of edges.

    All edges in a group, and functional groups held by them, should be
    used together in a single :class:`.Reaction`.

    """

    def __init__(self, edges):
        """
        Initialize an :class:`.EdgeGroup` instance.

        Parameters
        ----------
        edges : :class:`iterable` of :class:`.Edge`
            The edges in the group.

        """

        self._edge_ids = tuple(edge.get_id() for edge in edges)

    def get_edge_ids(self):
        """
        Yield the ids of edges in the group.

        Yields
        ------
        :class:`int`
            The id of an edge.

        """

        yield from self._edge_ids

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return f"{self.__class__.__name__}(edge_ids={self._edge_ids})"



class ConstructedMoleculeDatabase:
    """
    Abstract base class for storing constructed molecules.

    See Also
    --------
    :class:`.MoleculeDatabase`
        If you need to store and retrieve :class:`.Molecule`
        instances.

    Examples
    --------
    *Subclass Implementation*

    The source code of the subclasses, listed in
    :mod:`constructed_molecule_database \
    <~.databases.constructed_molecule>`, can
    serve as good examples.

    """

    def put(self, molecule):
        """
        Put `molecule` into the database.

        Parameters
        ----------
        molecule : :class:`.ConstructedMolecule`
            The molecule to place into the database.

        Returns
        -------
        None : :class:`NoneType`

        """

        raise NotImplementedError()

    def get(self, key):
        """
        Get the molecule with `key` from the database.

        Parameters
        ----------
        key : :class:`object`
            The key of a molecule, which is to be returned from the
            database.

        Returns
        -------
        :class:`.ConstructedMolecule`
            The molecule held in the database under `key`.

        Raises
        ------
        :class:`KeyError`
            If `key` is not found in the database.

        """

        raise NotImplementedError()

    def get_all(self):
        """
        Get all entries in the database.

        Yields
        ------
        :class:`.ConstructedMolecule`
            A molecule in the database.

        """

        raise NotImplementedError()



import pytest
import stk

from ..case_data import GenericCaseData


@pytest.fixture
def diol(get_atom_ids):
    a, b, c, d, e, f = get_atom_ids(6)
    return _diol(
        atom1=stk.C(a),
        oxygen1=stk.O(b),
        hydrogen1=stk.H(c),
        atom2=stk.C(d),
        oxygen2=stk.O(e),
        hydrogen2=stk.H(f),
    )


def _diol(atom1, oxygen1, hydrogen1, atom2, oxygen2, hydrogen2):
    bonders = (atom1, atom2)
    deleters = (oxygen1, hydrogen1, oxygen2, hydrogen2)
    return GenericCaseData(
        functional_group=stk.Diol(
            atom1=atom1,
            oxygen1=oxygen1,
            hydrogen1=hydrogen1,
            atom2=atom2,
            oxygen2=oxygen2,
            hydrogen2=hydrogen2,
            bonders=bonders,
            deleters=deleters,
        ),
        atoms=(atom1, oxygen1, hydrogen1, atom2, oxygen2, hydrogen2),
        bonders=bonders,
        deleters=deleters,
    )



from collections import defaultdict

import stk


def check_edge_functional_groups(
    old_state,
    new_state,
    building_blocks,
    placement_results,
):
    expected = get_expected_edge_functional_groups(
        old_state=old_state,
        building_blocks=building_blocks,
        placement_results=placement_results,
    )

    for edge_id in range(old_state.get_num_edges()):
        edge_group = stk.EdgeGroup((old_state.get_edge(edge_id),))
        expected_fgs = {get_fg_key(fg): fg for fg in expected[edge_id]}
        new_state_fgs = {
            get_fg_key(fg): fg
            for fg in new_state.get_edge_group_functional_groups(edge_group)
        }
        assert expected_fgs.keys() == new_state_fgs.keys()
        for fg_key in expected_fgs:
            is_equivalent_fg(
                fg1=expected_fgs[fg_key],
                fg2=new_state_fgs[fg_key],
            )


def get_expected_edge_functional_groups(
    old_state,
    building_blocks,
    placement_results,
):
    expected = defaultdict(list)
    for edge_id in range(old_state.get_num_edges()):
        edge_group = stk.EdgeGroup((old_state.get_edge(edge_id),))
        expected[edge_id].extend(
            old_state.get_edge_group_functional_groups(edge_group)
        )
    for edge_id, fg in added_functional_groups(
        old_state=old_state,
        building_blocks=building_blocks,
        placement_results=placement_results,
    ):
        expected[edge_id].append(fg)
    return expected


def is_equivalent_fg(fg1, fg2):
    assert fg1.__class__ is fg2.__class__
    assert tuple(fg1.get_atom_ids()) == tuple(fg2.get_atom_ids())
    assert tuple(fg1.get_placer_ids()) == tuple(fg2.get_placer_ids())


def get_fg_key(fg):
    return tuple(fg.get_atom_ids())


def added_functional_groups(
    old_state,
    building_blocks,
    placement_results,
):
    start_atom_id = sum(1 for _ in old_state.get_atoms())
    for building_block, result in zip(
        building_blocks,
        placement_results,
    ):
        atom_map = {
            atom.get_id(): atom.with_id(id_)
            for id_, atom in enumerate(
                building_block.get_atoms(),
                start_atom_id,
            )
        }
        start_atom_id += building_block.get_num_atoms()
        for fg_id, edge_id in result.functional_group_edges.items():
            fg = next(building_block.get_functional_groups(fg_id))
            yield edge_id, fg.with_atoms(atom_map)



class CaseData:
    """
    A :class:`.MolWriter` test case.

    Attributes
    ----------
    molecule : :class:`.Molecule`
        Molecule to test.

    writer : :class:`.MolWriter`
        The writer to test.

    string : :class:`str`
        The expected output string.

    """

    def __init__(self, molecule, writer, string):
        """
        Initialize a :class:`.CaseData` instance.

        Parameters
        ----------
        molecule : :class:`.Molecule`
            Molecule to test.

        writer : :class:`.MolWriter`
            The writer to test.

        string : :class:`str`
            The expected output string.

        """

        self.molecule = molecule
        self.writer = writer
        self.string = string



"""
M24L48
======

"""

from stk._internal.topology_graphs.edge import Edge

from .cage import Cage
from .vertices import LinearVertex, NonLinearVertex


class M24L48(Cage):
    """
    Represents a cage topology graph.

    .. moldoc::

        import moldoc.molecule as molecule
        import stk

        bb1 = stk.BuildingBlock(
            smiles='[Pd+2]',
            functional_groups=(
                stk.SingleAtom(stk.Pd(0, charge=2))
                for i in range(4)
            ),
            position_matrix=[[0, 0, 0]],
        )

        bb2 = stk.BuildingBlock(
            smiles=(
                'C1=NC=CC(C2=CC=CC(C3=C'
                'C=NC=C3)=C2)=C1'
            ),
            functional_groups=[
                stk.SmartsFunctionalGroupFactory(
                    smarts='[#6]~[#7X2]~[#6]',
                    bonders=(1, ),
                    deleters=(),
                ),
            ],
        )

        cage = stk.ConstructedMolecule(
            topology_graph=stk.cage.M24L48(
                building_blocks=(bb1, bb2),
            ),
        )

        moldoc_display_molecule = molecule.Molecule(
            atoms=(
                molecule.Atom(
                    atomic_number=atom.get_atomic_number(),
                    position=position,
                ) for atom, position in zip(
                    cage.get_atoms(),
                    cage.get_position_matrix(),
                )
            ),
            bonds=(
                molecule.Bond(
                    atom1_id=bond.get_atom1().get_id(),
                    atom2_id=bond.get_atom2().get_id(),
                    order=(
                        1
                        if bond.get_order() == 9
                        else bond.get_order()
                    ),
                ) for bond in cage.get_bonds()
            ),
        )

    Metal building blocks with four functional groups are
    required for this topology.

    Ligand building blocks with two functional groups are required for
    this topology.

    When using a :class:`dict` for the `building_blocks` parameter,
    as in :ref:`cage-topology-graph-examples`:
    *Multi-Building Block Cage Construction*, a
    :class:`.BuildingBlock`, with the following number of functional
    groups, needs to be assigned to each of the following vertex ids:

        | 4-functional groups: 0 to 23
        | 2-functional groups: 24 to 71

    See :class:`.Cage` for more details and examples.

    """

    _coord1 = 0.621
    _coord2 = -0.621
    _coord3 = _coord1 - (_coord1 - _coord2) / 2
    _coord4 = -1.5
    _coord5 = 1.5
    _coord6 = _coord5 - (_coord5 - _coord1) / 2

    _vertex_prototypes = (
        NonLinearVertex(0, [_coord1, _coord1, _coord4]),
        NonLinearVertex(1, [_coord1, _coord1, _coord5]),
        NonLinearVertex(2, [_coord1, _coord4, _coord2]),
        NonLinearVertex(3, [_coord1, _coord5, _coord2]),
        NonLinearVertex(4, [_coord1, _coord2, _coord4]),
        NonLinearVertex(5, [_coord1, _coord2, _coord5]),
        NonLinearVertex(6, [_coord1, _coord4, _coord1]),
        NonLinearVertex(7, [_coord1, _coord5, _coord1]),
        NonLinearVertex(8, [_coord2, _coord1, _coord4]),
        NonLinearVertex(9, [_coord2, _coord1, _coord5]),
        NonLinearVertex(10, [_coord2, _coord2, _coord4]),
        NonLinearVertex(11, [_coord2, _coord2, _coord5]),
        NonLinearVertex(12, [_coord2, _coord4, _coord1]),
        NonLinearVertex(13, [_coord2, _coord5, _coord1]),
        NonLinearVertex(14, [_coord2, _coord4, _coord2]),
        NonLinearVertex(15, [_coord2, _coord5, _coord2]),
        NonLinearVertex(16, [_coord4, _coord1, _coord1]),
        NonLinearVertex(17, [_coord4, _coord2, _coord1]),
        NonLinearVertex(18, [_coord4, _coord2, _coord2]),
        NonLinearVertex(19, [_coord4, _coord1, _coord2]),
        NonLinearVertex(20, [_coord5, _coord1, _coord1]),
        NonLinearVertex(21, [_coord5, _coord2, _coord1]),
        NonLinearVertex(22, [_coord5, _coord2, _coord2]),
        NonLinearVertex(23, [_coord5, _coord1, _coord2]),
        LinearVertex(24, [_coord1, _coord3, _coord4], False),
        LinearVertex(25, [_coord1, _coord4, _coord3], False),
        LinearVertex(26, [_coord1, _coord3, _coord5], False),
        LinearVertex(27, [_coord1, _coord5, _coord3], False),
        LinearVertex(28, [_coord2, _coord3, _coord4], False),
        LinearVertex(29, [_coord2, _coord4, _coord3], False),
        LinearVertex(30, [_coord2, _coord3, _coord5], False),
        LinearVertex(31, [_coord2, _coord5, _coord3], False),
        LinearVertex(32, [_coord3, _coord1, _coord4], False),
        LinearVertex(33, [_coord4, _coord1, _coord3], False),
        LinearVertex(34, [_coord3, _coord1, _coord5], False),
        LinearVertex(35, [_coord5, _coord1, _coord3], False),
        LinearVertex(36, [_coord3, _coord2, _coord4], False),
        LinearVertex(37, [_coord4, _coord2, _coord3], False),
        LinearVertex(38, [_coord3, _coord2, _coord5], False),
        LinearVertex(39, [_coord5, _coord2, _coord3], False),
        LinearVertex(40, [_coord3, _coord4, _coord1], False),
        LinearVertex(41, [_coord4, _coord3, _coord1], False),
        LinearVertex(42, [_coord3, _coord5, _coord1], False),
        LinearVertex(43, [_coord5, _coord3, _coord1], False),
        LinearVertex(44, [_coord3, _coord4, _coord2], False),
        LinearVertex(45, [_coord4, _coord3, _coord2], False),
        LinearVertex(46, [_coord3, _coord5, _coord2], False),
        LinearVertex(47, [_coord5, _coord3, _coord2], False),
        LinearVertex(48, [_coord1, _coord6, _coord6], False),
        LinearVertex(49, [_coord1, _coord6, -_coord6], False),
        LinearVertex(50, [_coord1, -_coord6, _coord6], False),
        LinearVertex(51, [_coord1, -_coord6, -_coord6], False),
        LinearVertex(52, [_coord2, _coord6, _coord6], False),
        LinearVertex(53, [_coord2, _coord6, -_coord6], False),
        LinearVertex(54, [_coord2, -_coord6, _coord6], False),
        LinearVertex(55, [_coord2, -_coord6, -_coord6], False),
        LinearVertex(56, [_coord6, _coord1, _coord6], False),
        LinearVertex(57, [_coord6, _coord1, -_coord6], False),
        LinearVertex(58, [-_coord6, _coord1, _coord6], False),
        LinearVertex(59, [-_coord6, _coord1, -_coord6], False),
        LinearVertex(60, [_coord6, _coord2, _coord6], False),
        LinearVertex(61, [_coord6, _coord2, -_coord6], False),
        LinearVertex(62, [-_coord6, _coord2, _coord6], False),
        LinearVertex(63, [-_coord6, _coord2, -_coord6], False),
        LinearVertex(64, [_coord6, _coord6, _coord1], False),
        LinearVertex(65, [_coord6, -_coord6, _coord1], False),
        LinearVertex(66, [-_coord6, _coord6, _coord1], False),
        LinearVertex(67, [-_coord6, -_coord6, _coord1], False),
        LinearVertex(68, [_coord6, _coord6, _coord2], False),
        LinearVertex(69, [_coord6, -_coord6, _coord2], False),
        LinearVertex(70, [-_coord6, _coord6, _coord2], False),
        LinearVertex(71, [-_coord6, -_coord6, _coord2], False),
    )

    _edge_prototypes = (
        Edge(0, _vertex_prototypes[0], _vertex_prototypes[57]),
        Edge(1, _vertex_prototypes[0], _vertex_prototypes[32]),
        Edge(2, _vertex_prototypes[0], _vertex_prototypes[49]),
        Edge(3, _vertex_prototypes[0], _vertex_prototypes[24]),
        Edge(4, _vertex_prototypes[1], _vertex_prototypes[26]),
        Edge(5, _vertex_prototypes[1], _vertex_prototypes[56]),
        Edge(6, _vertex_prototypes[1], _vertex_prototypes[48]),
        Edge(7, _vertex_prototypes[1], _vertex_prototypes[34]),
        Edge(8, _vertex_prototypes[2], _vertex_prototypes[44]),
        Edge(9, _vertex_prototypes[2], _vertex_prototypes[25]),
        Edge(10, _vertex_prototypes[2], _vertex_prototypes[69]),
        Edge(11, _vertex_prototypes[2], _vertex_prototypes[51]),
        Edge(12, _vertex_prototypes[3], _vertex_prototypes[68]),
        Edge(13, _vertex_prototypes[3], _vertex_prototypes[49]),
        Edge(14, _vertex_prototypes[3], _vertex_prototypes[27]),
        Edge(15, _vertex_prototypes[3], _vertex_prototypes[46]),
        Edge(16, _vertex_prototypes[4], _vertex_prototypes[51]),
        Edge(17, _vertex_prototypes[4], _vertex_prototypes[36]),
        Edge(18, _vertex_prototypes[4], _vertex_prototypes[24]),
        Edge(19, _vertex_prototypes[4], _vertex_prototypes[61]),
        Edge(20, _vertex_prototypes[5], _vertex_prototypes[50]),
        Edge(21, _vertex_prototypes[5], _vertex_prototypes[60]),
        Edge(22, _vertex_prototypes[5], _vertex_prototypes[26]),
        Edge(23, _vertex_prototypes[5], _vertex_prototypes[38]),
        Edge(24, _vertex_prototypes[6], _vertex_prototypes[40]),
        Edge(25, _vertex_prototypes[6], _vertex_prototypes[25]),
        Edge(26, _vertex_prototypes[6], _vertex_prototypes[65]),
        Edge(27, _vertex_prototypes[6], _vertex_prototypes[50]),
        Edge(28, _vertex_prototypes[7], _vertex_prototypes[64]),
        Edge(29, _vertex_prototypes[7], _vertex_prototypes[48]),
        Edge(30, _vertex_prototypes[7], _vertex_prototypes[27]),
        Edge(31, _vertex_prototypes[7], _vertex_prototypes[42]),
        Edge(32, _vertex_prototypes[8], _vertex_prototypes[53]),
        Edge(33, _vertex_prototypes[8], _vertex_prototypes[32]),
        Edge(34, _vertex_prototypes[8], _vertex_prototypes[28]),
        Edge(35, _vertex_prototypes[8], _vertex_prototypes[59]),
        Edge(36, _vertex_prototypes[9], _vertex_prototypes[34]),
        Edge(37, _vertex_prototypes[9], _vertex_prototypes[52]),
        Edge(38, _vertex_prototypes[9], _vertex_prototypes[30]),
        Edge(39, _vertex_prototypes[9], _vertex_prototypes[58]),
        Edge(40, _vertex_prototypes[10], _vertex_prototypes[63]),
        Edge(41, _vertex_prototypes[10], _vertex_prototypes[28]),
        Edge(42, _vertex_prototypes[10], _vertex_prototypes[36]),
        Edge(43, _vertex_prototypes[10], _vertex_prototypes[55]),
        Edge(44, _vertex_prototypes[11], _vertex_prototypes[38]),
        Edge(45, _vertex_prototypes[11], _vertex_prototypes[54]),
        Edge(46, _vertex_prototypes[11], _vertex_prototypes[62]),
        Edge(47, _vertex_prototypes[11], _vertex_prototypes[30]),
        Edge(48, _vertex_prototypes[12], _vertex_prototypes[67]),
        Edge(49, _vertex_prototypes[12], _vertex_prototypes[54]),
        Edge(50, _vertex_prototypes[12], _vertex_prototypes[29]),
        Edge(51, _vertex_prototypes[12], _vertex_prototypes[40]),
        Edge(52, _vertex_prototypes[13], _vertex_prototypes[42]),
        Edge(53, _vertex_prototypes[13], _vertex_prototypes[31]),
        Edge(54, _vertex_prototypes[13], _vertex_prototypes[66]),
        Edge(55, _vertex_prototypes[13], _vertex_prototypes[52]),
        Edge(56, _vertex_prototypes[14], _vertex_prototypes[71]),
        Edge(57, _vertex_prototypes[14], _vertex_prototypes[55]),
        Edge(58, _vertex_prototypes[14], _vertex_prototypes[44]),
        Edge(59, _vertex_prototypes[14], _vertex_prototypes[29]),
        Edge(60, _vertex_prototypes[15], _vertex_prototypes[46]),
        Edge(61, _vertex_prototypes[15], _vertex_prototypes[31]),
        Edge(62, _vertex_prototypes[15], _vertex_prototypes[70]),
        Edge(63, _vertex_prototypes[15], _vertex_prototypes[53]),
        Edge(64, _vertex_prototypes[16], _vertex_prototypes[66]),
        Edge(65, _vertex_prototypes[16], _vertex_prototypes[58]),
        Edge(66, _vertex_prototypes[16], _vertex_prototypes[41]),
        Edge(67, _vertex_prototypes[16], _vertex_prototypes[33]),
        Edge(68, _vertex_prototypes[17], _vertex_prototypes[41]),
        Edge(69, _vertex_prototypes[17], _vertex_prototypes[37]),
        Edge(70, _vertex_prototypes[17], _vertex_prototypes[67]),
        Edge(71, _vertex_prototypes[17], _vertex_prototypes[62]),
        Edge(72, _vertex_prototypes[18], _vertex_prototypes[45]),
        Edge(73, _vertex_prototypes[18], _vertex_prototypes[37]),
        Edge(74, _vertex_prototypes[18], _vertex_prototypes[71]),
        Edge(75, _vertex_prototypes[18], _vertex_prototypes[63]),
        Edge(76, _vertex_prototypes[19], _vertex_prototypes[70]),
        Edge(77, _vertex_prototypes[19], _vertex_prototypes[59]),
        Edge(78, _vertex_prototypes[19], _vertex_prototypes[45]),
        Edge(79, _vertex_prototypes[19], _vertex_prototypes[33]),
        Edge(80, _vertex_prototypes[20], _vertex_prototypes[43]),
        Edge(81, _vertex_prototypes[20], _vertex_prototypes[35]),
        Edge(82, _vertex_prototypes[20], _vertex_prototypes[56]),
        Edge(83, _vertex_prototypes[20], _vertex_prototypes[64]),
        Edge(84, _vertex_prototypes[21], _vertex_prototypes[43]),
        Edge(85, _vertex_prototypes[21], _vertex_prototypes[39]),
        Edge(86, _vertex_prototypes[21], _vertex_prototypes[65]),
        Edge(87, _vertex_prototypes[21], _vertex_prototypes[60]),
        Edge(88, _vertex_prototypes[22], _vertex_prototypes[69]),
        Edge(89, _vertex_prototypes[22], _vertex_prototypes[61]),
        Edge(90, _vertex_prototypes[22], _vertex_prototypes[47]),
        Edge(91, _vertex_prototypes[22], _vertex_prototypes[39]),
        Edge(92, _vertex_prototypes[23], _vertex_prototypes[47]),
        Edge(93, _vertex_prototypes[23], _vertex_prototypes[57]),
        Edge(94, _vertex_prototypes[23], _vertex_prototypes[68]),
        Edge(95, _vertex_prototypes[23], _vertex_prototypes[35]),
    )

    _num_windows = 26
    _num_window_types = 2



import pytest
import stk

from .case_data import CaseData


@pytest.fixture(params=[stk.C(0), stk.H(1), stk.K(12)])
def atom(request):
    """
    An :class:`.Atom` instance.

    """

    return request.param.clone()


@pytest.fixture
def atom1(atom):
    """
    An :class:`.Atom` instance.

    """

    return atom.clone()


@pytest.fixture
def atom2(atom):
    """
    An :class:`.Atom` instance.

    """

    return atom.clone()


@pytest.fixture(params=[1, 2, 3, 4, 1.5])
def order(request):
    """
    A possible bond order.

    """

    return request.param


@pytest.fixture(params=list(range(-3, 4)))
def periodicity_value(request):
    """
    The periodicity of a bond along 1 axis.

    """

    return request.param


@pytest.fixture
def x_periodicity(periodicity_value):
    """
    The periodicity of a bond along the x axis.

    """

    return periodicity_value


@pytest.fixture
def y_periodicity(periodicity_value):
    """
    The periodicity of a bond along the y axis.

    """

    return periodicity_value


@pytest.fixture
def z_periodicity(periodicity_value):
    """
    The periodicity of a bond along the z axis.

    """

    return periodicity_value


@pytest.fixture
def periodicity(x_periodicity, y_periodicity, z_periodicity):
    """
    A possible bond periodicity.

    """

    return x_periodicity, y_periodicity, z_periodicity


@pytest.fixture
def case_data(atom1, atom2, order, periodicity):
    """
    A :class:`.CaseData` instance.

    """

    return CaseData(
        bond=stk.Bond(atom1, atom2, order, periodicity),
        atom1=atom1,
        atom2=atom2,
        order=order,
        periodicity=periodicity,
    )


@pytest.fixture
def bond(case_data):
    """
    A :class:`.Bond` instance.

    """

    return case_data.bond



def test_to_rdkit_mol(molecule):
    """
    Test :meth:`.Molecule.to_rdkit_mol`.

    Parameters
    ----------
    molecule : :class:`.Molecule`
        The molecule to test.

    Returns
    -------
    None : :class:`NoneType`

    """

    rdkit_molecule = molecule.to_rdkit_mol()
    assert rdkit_molecule.GetNumConformers() == 1

    assert rdkit_molecule.GetNumAtoms() == molecule.get_num_atoms()
    atoms = zip(molecule.get_atoms(), rdkit_molecule.GetAtoms())
    for atom, rdkit_atom in atoms:
        assert atom.get_charge() == rdkit_atom.GetFormalCharge()
        assert atom.get_atomic_number() == rdkit_atom.GetAtomicNum()

    assert molecule.get_num_bonds() == rdkit_molecule.GetNumBonds()
    bonds = zip(molecule.get_bonds(), rdkit_molecule.GetBonds())
    for bond, rdkit_bond in bonds:
        assert bond.get_order() == rdkit_bond.GetBondTypeAsDouble()
        assert bond.get_atom1().get_id() == rdkit_bond.GetBeginAtomIdx()
        assert bond.get_atom2().get_id() == rdkit_bond.GetEndAtomIdx()



import pymongo
import pytest
import stk

from ..case_data import CaseData
from .utilities import Counter

_counter = Counter()


def _get_case_data(mongo_client: pymongo.MongoClient) -> CaseData:
    # The basic idea here is that the _counter.get_count method will
    # return a different "fitness value" each time it is called.
    # When the test runs fitness_calculator.get_fitness_value(), if
    # caching is working, the same number as before will be returned.
    # However, if caching is not working, a different number will be
    # returned as the fitness value.

    db = stk.ValueMongoDb(
        mongo_client=mongo_client,
        collection="test_caching",
        database="_stk_pytest_database",
    )

    fitness_calculator = stk.PropertyVector(
        property_functions=(_counter.get_count,),
        input_database=db,
        output_database=db,
    )
    record = stk.MoleculeRecord(
        topology_graph=stk.polymer.Linear(
            building_blocks=(stk.BuildingBlock("BrCCBr"),),
            repeating_unit="A",
            num_repeating_units=1,
        ),
    )
    fitness_value = fitness_calculator.get_fitness_value(record)

    return CaseData(
        fitness_calculator=fitness_calculator,
        record=record,
        fitness_value=fitness_value,
    )


@pytest.fixture(
    params=(_get_case_data,),
)
def property_vector(
    request: pytest.FixtureRequest,
    mongo_client: pymongo.MongoClient,
) -> CaseData:
    return request.param(mongo_client)



def check_vertices(old_state, new_state):
    assert old_state.get_num_vertices() == new_state.get_num_vertices()
    for vertex_id in range(old_state.get_num_vertices()):
        assert old_state.get_vertex(vertex_id) is new_state.get_vertex(
            vertex_id
        )



import numpy as np

from ...utilities import normalize_ids
from ..utilities import get_centroid


def test_get_centroid(case_data, get_atom_ids):
    """
    Test :meth:`.Molecule.get_centroid`.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        A test case. Holds the molecule to test and the correct atom
        positions.

    get_atom_ids : :class:`callable`
        Takes a single parameter, `molecule`, and returns a valid
        `atom_ids` parameter for :meth:`.Molecule.get_centroid`. This
        allows the testing of different values of this parameter.

    Returns
    -------
    None : :class:`NoneType`

    Notes
    -----
    This test compares the result of :meth:`.Molecule.get_centroid`
    to the result of :func:`.get_centroid`, which is a utility
    function defined for the purposes of this test. Because
    :func:`.get_centroid` is tested independently, in
    :mod:`.test_get_centroid_helper`, if its tests pass, then
    it can be assumed, that :func:`.get_centroid` gives correct
    results.

    Now, assuming that :func:`.get_centroid` passed all of its tests,
    this test compares the results of :meth:`.Molecule.get_centroid`
    to the results of :func:`.get_centroid`. If the results do not
    match, the fault can be placed on :meth:`.Molecule.get_centroid`,
    because :func:`.get_centroid` has already been verified to be
    correct by its own tests.

    """

    _test_get_centroid(
        molecule=case_data.molecule,
        centroid=get_centroid(
            position_matrix=case_data.position_matrix,
            atom_ids=tuple(
                normalize_ids(
                    molecule=case_data.molecule,
                    ids=get_atom_ids(case_data.molecule),
                )
            ),
        ),
        get_atom_ids=get_atom_ids,
    )


def _test_get_centroid(molecule, centroid, get_atom_ids):
    """
    Test :meth:`.Molecule.get_centroid`.

    Parameters
    ----------
    molecule : :class:`.Molecule`
        The molecule to test.

    centroid : :class:`numpy.ndarray`
        The correct centroid of `molecule`.

    get_atom_ids : :class:`callable`
        Takes a single parameter, `molecule`, and returns a valid
        `atom_ids` parameter for :meth:`.Molecule.get_centroid`. This
        allows the testing of different values of this parameter.

    Returns
    -------
    None : :class:`NoneType`

    """

    assert np.allclose(
        a=centroid,
        b=molecule.get_centroid(get_atom_ids(molecule)),
        atol=1e-6,
    )



from __future__ import annotations

import typing
from collections import abc

from stk._internal.building_block import BuildingBlock
from stk._internal.construction_state import (
    ConstructionState,
)
from stk._internal.optimizers.null import (
    NullOptimizer,
)
from stk._internal.optimizers.optimizer import (
    Optimizer,
)
from stk._internal.reaction_factories.generic_reaction_factory import (
    GenericReactionFactory,
)
from stk._internal.topology_graphs.topology_graph.topology_graph import (
    TopologyGraph,
)
from stk._internal.topology_graphs.vertex import (
    Vertex,
)

from .vertices import GuestVertex, HostVertex


class Guest:
    """
    Holds the data defining the placement of a guest molecule.

    """

    def __init__(
        self,
        building_block: BuildingBlock,
        start_vector: tuple[float, float, float] = (1.0, 0.0, 0.0),
        end_vector: tuple[float, float, float] = (1.0, 0.0, 0.0),
        displacement: tuple[float, float, float] = (1.0, 0.0, 0.0),
    ) -> None:
        """
        Initialize a :class:`.Guest` instance.

        Parameters:
            building_block: The guest molecule.

            start_vector: A direction vector which gets aligned with
                `end_vector`.

            end_vector: A direction vector which determines the
                rotation applied to the `building_block`. A rotation
                such that `start_vector` is transformed into
                `end_vector` is applied.

            displacement: The translational offset of the guest.

        """

        self._building_block = building_block
        self._start_vector = start_vector
        self._end_vector = end_vector
        self._displacement = displacement

    def get_building_block(self) -> BuildingBlock:
        """
        Return the building block.

        Returns:
            The building block.

        """

        return self._building_block

    def get_start_vector(self) -> tuple[float, float, float]:
        """
        Return the start vector.

        Returns:
            The start vector.

        """

        return self._start_vector

    def get_end_vector(self) -> tuple[float, float, float]:
        """
        Return the end vector.

        Returns:
            The end vector.

        """

        return self._end_vector

    def get_displacement(self) -> tuple[float, float, float]:
        """
        Return the displacement.

        Returns:
            The displacement.

        """

        return self._displacement

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"{self._building_block!r}, "
            f"start_vector={self._start_vector!r}, "
            f"end_vector={self._end_vector!r}, "
            f"displacement={self._displacement!r})"
        )


class Complex(TopologyGraph):
    """
    Represents a host-guest complex topology graph.

    Host and guest building blocks do not require functional groups.

    Examples:

        *Construction*

        You can use :class:`.ConstructedMolecule` instances as the
        host, but you should turn them into a :class:`.BuildingBlock`
        first

        .. testcode:: construction

            import stk

            host = stk.ConstructedMolecule(
                topology_graph=stk.cage.FourPlusSix(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles='NC1CCCCC1N',
                            functional_groups=[
                                stk.PrimaryAminoFactory(),
                            ],
                        ),
                        stk.BuildingBlock(
                            smiles='O=Cc1cc(C=O)cc(C=O)c1',
                            functional_groups=[stk.AldehydeFactory()],
                        ),
                    ),
                    optimizer=stk.MCHammer(),
                ),
            )
            complex = stk.ConstructedMolecule(
                topology_graph=stk.host_guest.Complex(
                    host=stk.BuildingBlock.init_from_molecule(host),
                    guests=stk.host_guest.Guest(
                        building_block=stk.BuildingBlock('[Br][Br]'),
                    ),
                ),
            )

        .. moldoc::

            import moldoc.molecule as molecule
            import stk

            host = stk.ConstructedMolecule(
                topology_graph=stk.cage.FourPlusSix(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles='NC1CCCCC1N',
                            functional_groups=[
                                stk.PrimaryAminoFactory(),
                            ],
                        ),
                        stk.BuildingBlock(
                            smiles='O=Cc1cc(C=O)cc(C=O)c1',
                            functional_groups=[stk.AldehydeFactory()],
                        ),
                    ),
                    optimizer=stk.MCHammer(),
                ),
            )
            complex = stk.ConstructedMolecule(
                topology_graph=stk.host_guest.Complex(
                    host=stk.BuildingBlock.init_from_molecule(host),
                    guests=stk.host_guest.Guest(
                        building_block=stk.BuildingBlock('[Br][Br]'),
                    ),
                ),
            )

            moldoc_display_molecule = molecule.Molecule(
                atoms=(
                    molecule.Atom(
                        atomic_number=atom.get_atomic_number(),
                        position=position,
                    ) for atom, position in zip(
                        complex.get_atoms(),
                        complex.get_position_matrix(),
                    )
                ),
                bonds=(
                    molecule.Bond(
                        atom1_id=bond.get_atom1().get_id(),
                        atom2_id=bond.get_atom2().get_id(),
                        order=bond.get_order(),
                    ) for bond in complex.get_bonds()
                ),
            )

        You can also generate complexes with multiple guests.

        .. testcode:: multi-guest-construction

            import stk

            host = stk.ConstructedMolecule(
                topology_graph=stk.cage.FourPlusSix(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles='NC1CCCCC1N',
                            functional_groups=[
                                stk.PrimaryAminoFactory(),
                            ],
                        ),
                        stk.BuildingBlock(
                            smiles='O=Cc1cc(C=O)cc(C=O)c1',
                            functional_groups=[stk.AldehydeFactory()],
                        ),
                    ),
                    optimizer=stk.MCHammer(),
                ),
            )
            guest1 = stk.host_guest.Guest(
                building_block=stk.BuildingBlock('BrBr'),
                displacement=(0., 3., 0.),
            )
            guest2 = stk.host_guest.Guest(
                building_block=stk.BuildingBlock('C1CCCC1'),
            )

            complex = stk.ConstructedMolecule(
                topology_graph=stk.host_guest.Complex(
                    host=stk.BuildingBlock.init_from_molecule(host),
                    guests=(guest1, guest2),
                ),
            )

        .. moldoc::

            import moldoc.molecule as molecule
            import stk

            host = stk.ConstructedMolecule(
                topology_graph=stk.cage.FourPlusSix(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles='NC1CCCCC1N',
                            functional_groups=[
                                stk.PrimaryAminoFactory(),
                            ],
                        ),
                        stk.BuildingBlock(
                            smiles='O=Cc1cc(C=O)cc(C=O)c1',
                            functional_groups=[stk.AldehydeFactory()],
                        ),
                    ),
                    optimizer=stk.MCHammer(),
                ),
            )
            guest1 = stk.host_guest.Guest(
                building_block=stk.BuildingBlock('BrBr'),
                displacement=(0., 3., 0.),
            )
            guest2 = stk.host_guest.Guest(
                building_block=stk.BuildingBlock('C1CCCC1'),
            )

            complex = stk.ConstructedMolecule(
                topology_graph=stk.host_guest.Complex(
                    host=stk.BuildingBlock.init_from_molecule(host),
                    guests=(guest1, guest2),
                ),
            )

            moldoc_display_molecule = molecule.Molecule(
                atoms=(
                    molecule.Atom(
                        atomic_number=atom.get_atomic_number(),
                        position=position,
                    ) for atom, position in zip(
                        complex.get_atoms(),
                        complex.get_position_matrix(),
                    )
                ),
                bonds=(
                    molecule.Bond(
                        atom1_id=bond.get_atom1().get_id(),
                        atom2_id=bond.get_atom2().get_id(),
                        order=bond.get_order(),
                    ) for bond in complex.get_bonds()
                ),
            )

        *Suggested Optimization*

        For :class:`.Complex` topologies, it is recommended to use the
        :class:`.Spinner` optimizer. It is also recommended that the
        building blocks are already optimized prior to construction.
        This optimizer will work on multi-guest systems.

        .. testcode:: suggested-optimization

            import stk

            host = stk.ConstructedMolecule(
                topology_graph=stk.cage.FourPlusSix(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles='NC1CCCCC1N',
                            functional_groups=[
                                stk.PrimaryAminoFactory(),
                            ],
                        ),
                        stk.BuildingBlock(
                            smiles='O=Cc1cc(C=O)cc(C=O)c1',
                            functional_groups=[stk.AldehydeFactory()],
                        ),
                    ),
                    optimizer=stk.MCHammer(),
                ),
            )
            guest1 = stk.host_guest.Guest(
                building_block=stk.BuildingBlock('BrBr'),
                displacement=(0., 3., 0.),
            )
            guest2 = stk.host_guest.Guest(
                building_block=stk.BuildingBlock('C1CCCC1'),
            )

            complex = stk.ConstructedMolecule(
                topology_graph=stk.host_guest.Complex(
                    host=stk.BuildingBlock.init_from_molecule(host),
                    guests=(guest1, guest2),
                    optimizer=stk.Spinner(),
                ),
            )

        .. moldoc::

            import moldoc.molecule as molecule
            import stk

            host = stk.ConstructedMolecule(
                topology_graph=stk.cage.FourPlusSix(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles='NC1CCCCC1N',
                            functional_groups=[
                                stk.PrimaryAminoFactory(),
                            ],
                        ),
                        stk.BuildingBlock(
                            smiles='O=Cc1cc(C=O)cc(C=O)c1',
                            functional_groups=[stk.AldehydeFactory()],
                        ),
                    ),
                    optimizer=stk.MCHammer(),
                ),
            )
            guest1 = stk.host_guest.Guest(
                building_block=stk.BuildingBlock('BrBr'),
                displacement=(0., 3., 0.),
            )
            guest2 = stk.host_guest.Guest(
                building_block=stk.BuildingBlock('C1CCCC1'),
            )

            complex = stk.ConstructedMolecule(
                topology_graph=stk.host_guest.Complex(
                    host=stk.BuildingBlock.init_from_molecule(host),
                    guests=(guest1, guest2),
                    optimizer=stk.Spinner(),
                ),
            )

            moldoc_display_molecule = molecule.Molecule(
                atoms=(
                    molecule.Atom(
                        atomic_number=atom.get_atomic_number(),
                        position=position,
                    ) for atom, position in zip(
                        complex.get_atoms(),
                        complex.get_position_matrix(),
                    )
                ),
                bonds=(
                    molecule.Bond(
                        atom1_id=bond.get_atom1().get_id(),
                        atom2_id=bond.get_atom2().get_id(),
                        order=bond.get_order(),
                    ) for bond in complex.get_bonds()
                ),
            )

        *Changing the Position of the Guest*

        You can change the position and orientation of the guest, as
        well as its displacement

        .. testcode:: changing-the-position-of-the-guest

            import stk

            host = stk.ConstructedMolecule(
                topology_graph=stk.cage.FourPlusSix(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles='BrCCBr',
                            functional_groups=[stk.BromoFactory()],
                        ),
                        stk.BuildingBlock(
                            smiles='BrCC(Br)CBr',
                            functional_groups=[stk.BromoFactory()],
                        ),
                    ),
                ),
            )

            guest_building_block = stk.BuildingBlock('[Br][Br]')
            guest = stk.host_guest.Guest(
                building_block=guest_building_block,
                # Apply a rotation onto the guest molecule such that
                # the vector returned by get_direction() has the same
                # direction as [1, 1, 1].
                start_vector=guest_building_block.get_direction(),
                end_vector=[1, 1, 1],
                # Change the displacement of the guest.
                displacement=[5.3, 2.1, 7.1],
            )
            complex = stk.ConstructedMolecule(
                topology_graph=stk.host_guest.Complex(
                    host=stk.BuildingBlock.init_from_molecule(host),
                    guests=guest,
                ),
            )

    """

    def __init__(
        self,
        host: BuildingBlock,
        guests: typing.Union[Guest, typing.Iterable[Guest]],
        num_processes: int = 1,
        optimizer: Optimizer = NullOptimizer(),
    ) -> None:
        """
        Initialize an instance of :class:`.Complex`.

        Parameters:
            host: The host molecule.

            guests: The guest molecules. Can be a single
                :class:`.Guest` instance if only one guest is being
                used.

            num_processes: The number of parallel processes to create
                during :meth:`construct`.

            optimizer: Used to optimize the structure of the
                constructed molecule.

        """

        building_block_vertices = self._get_vertices_from_guests(
            host=host,
            guests=guests,
        )

        super().__init__(
            building_block_vertices=building_block_vertices,
            edges=(),
            reaction_factory=GenericReactionFactory(),
            construction_stages=(),
            num_processes=num_processes,
            optimizer=optimizer,
            edge_groups=(),
        )

    def _get_vertices_from_guests(
        self,
        host: BuildingBlock,
        guests: typing.Union[Guest, typing.Iterable[Guest]],
    ) -> dict[BuildingBlock, abc.Sequence[Vertex]]:
        if isinstance(guests, Guest):
            guests = (guests,)

        building_block_vertices: dict[BuildingBlock, abc.Sequence[Vertex]]
        building_block_vertices = {host: (HostVertex(0, (0.0, 0.0, 0.0)),)}
        guest_vertices = {
            guest.get_building_block(): (
                GuestVertex(
                    id=i + 1,
                    position=guest.get_displacement(),
                    start=guest.get_start_vector(),
                    target=guest.get_end_vector(),
                ),
            )
            for i, guest in enumerate(guests)
        }
        building_block_vertices.update(guest_vertices)

        return building_block_vertices

    def clone(self) -> Complex:
        return self._clone()

    def _run_reactions(
        self,
        state: ConstructionState,
    ) -> ConstructionState:
        return state

    def _get_scale(
        self,
        building_block_vertices: dict[BuildingBlock, abc.Sequence[Vertex]],
    ) -> float:
        return 1.0

    def __repr__(self) -> str:
        return "host_guest.Complex()"



from __future__ import annotations

import numpy as np
import pytest
import stk
from pytest_lazyfixture import lazy_fixture

from ...case_data import CaseData


@pytest.fixture(
    params=(
        lazy_fixture("cage1"),
        lazy_fixture("cage2"),
        lazy_fixture("cage3"),
    ),
)
def case_data(request):
    return request.param


@pytest.fixture
def cage1(cls, position):
    return CaseData(
        vertex=cls(
            id=0,
            position=position,
        ),
        id=0,
        position=position,
        cell=np.array([0, 0, 0]),
    )


@pytest.fixture
def cage2(init_at_center, vertices_):
    return CaseData(
        vertex=init_at_center(
            id=0,
            vertices=vertices_,
        ),
        id=0,
        position=(sum(v.get_position() for v in vertices_) / len(vertices_)),
        cell=np.array([0, 0, 0]),
    )


@pytest.fixture
def cage3(position):
    return CaseData(
        vertex=stk.cage.UnaligningVertex(
            id=0,
            position=position,
        ),
        id=0,
        position=position,
        cell=np.array([0, 0, 0]),
    )


@pytest.fixture(
    params=(
        [0, 0, 0],
        [1, 2, -20],
    ),
)
def position(request):
    return np.array(request.param, dtype=np.float64)


@pytest.fixture(
    params=(
        stk.cage.LinearVertex,
        stk.cage.NonLinearVertex,
        stk.cage.UnaligningVertex,
        stk.cage.AngledVertex,
    ),
)
def cls(request) -> stk.Vertex:
    return request.param


@pytest.fixture(
    params=(
        stk.cage.LinearVertex.init_at_center,
        stk.cage.NonLinearVertex.init_at_center,
        stk.cage.UnaligningVertex.init_at_center,
        stk.cage.AngledVertex.init_at_center,
    ),
)
def init_at_center(request) -> stk.Vertex:
    return request.param


@pytest.fixture(
    params=(
        lambda: (stk.Vertex(0, [1, 2, 3]),),
        lambda: (
            stk.Vertex(0, [1, 2, 3]),
            stk.Vertex(1, [-1, 2, -32]),
        ),
    ),
)
def vertices_(request) -> tuple[stk.Vertex, ...]:
    return request.param()



"""
Periodic Collapser
==================

"""

import mchammer as mch

from .optimizer import Optimizer
from .utilities import get_long_bond_ids, get_mch_bonds, get_subunits


class PeriodicCollapser(Optimizer):
    """
    Performs rigid-body collapse of molecules.

    This :class:`.Optimizer` will also update the `.PeriodicInfo`.

    Examples
    --------
    *Structure Optimization*

    Using :class:`.PeriodicCollapser` will lead to
    :class:`.ConstructedMolecule` structures without long bonds and
    match the unit-cell to the new structure.

    .. testcode:: structure-optimization

        import stk

        bb1 = stk.BuildingBlock('BrCCBr', [stk.BromoFactory()])
        bb2 = stk.BuildingBlock('BrCC(CBr)CBr', [stk.BromoFactory()])

        topology_graph = stk.cof.PeriodicHoneycomb(
            building_blocks=(bb1, bb2),
            lattice_size=(1, 2, 3),
            optimizer=stk.PeriodicCollapser(),
        )
        cof = stk.ConstructedMolecule(topology_graph)

    Optimisation with :mod:`stk` simply collects the final position
    matrix and periodic info. The optimisation's trajectory can be
    output using the :mod:`MCHammer` implementation if required by the
    user [#mchammer]_.

    The open-source optimization code :mod:`MCHammer` specializes in
    the `collapsing` of molecules with long bonds like those
    constructed by :mod:`stk`. This code is entirely nonphysical and
    is, therefore, completely general to any chemistry.

    References
    ----------
    .. [#mchammer] https://github.com/andrewtarzia/MCHammer

    """

    def __init__(
        self,
        step_size=0.1,
        distance_threshold=1.5,
        scale_steps=False,
    ):
        """
        Initialize an instance of :class:`.PeriodicCollapser`.

        Parameters
        ----------
        step_size : :class:`float`, optional
            The relative size of the step to take during collapse in
            Angstrom.

        distance_threshold : :class:`float`, optional
            Distance between distinct building blocks to use as
            threshold for halting collapse in Angstrom.

        scale_steps : :class:`bool`, optional
            Whether to scale the step of each distinct building block
            by its relative distance from the molecules centroid.

        """

        self._optimizer = mch.Collapser(
            step_size=step_size,
            distance_threshold=distance_threshold,
            scale_steps=scale_steps,
        )

    def optimize(self, state):
        mch_mol = mch.Molecule(
            atoms=(
                mch.Atom(
                    id=atom.get_id(),
                    element_string=atom.__class__.__name__,
                )
                for atom in state.get_atoms()
            ),
            bonds=get_mch_bonds(state),
            position_matrix=state.get_position_matrix(),
        )

        mch_mol, result = self._optimizer.get_result(
            mol=mch_mol,
            bond_pair_ids=tuple(get_long_bond_ids(state)),
            subunits=get_subunits(state),
        )

        old_pos_mat = state.get_position_matrix()
        new_pos_mat = mch_mol.get_position_matrix()
        old_extents = (
            abs(max(old_pos_mat[:, i]) - min(old_pos_mat[:, i]))
            for i in range(3)
        )
        new_extents = (
            abs(max(new_pos_mat[:, i]) - min(new_pos_mat[:, i]))
            for i in range(3)
        )
        ratios = (n / o for n, o in zip(new_extents, old_extents))
        old_lattice = state.get_lattice_constants()
        new_lattice = tuple(
            old_lattice[i] * ratio for i, ratio in enumerate(ratios)
        )
        state = state.with_lattice_constants(new_lattice)
        return state.with_position_matrix(
            position_matrix=mch_mol.get_position_matrix()
        )



class ReactionFactory:
    """
    An abstract base class for reaction factories.

    Reaction factories are responsible for creating :class:`.Reaction`
    instances. Different subclasses of this class will provide
    different options for creating :class:`.Reaction` instances.

    If you want to change which reactions are used to create and
    delete atoms and bonds during :class:`.ConstructedMolecule`
    construction, you want to subclass this abstract base class
    and implement :meth:`.get_reaction`. Your implementation can
    then pick which reaction to use for a particular
    :class:`.EdgeGroup`. You will then pass an instance of your
    :class:`.ReactionFactory` subclass to the :class:`.TopologyGraph`
    initializer, so that it knows to use it.

    Examples
    --------
    *Subclass Implementation*

    The source code of the subclasses, listed in
    :mod:`.reaction_factory`, can serve as good examples.

    """

    def get_reaction(self, construction_state, edge_group):
        """
        Get a reaction to use on the `edge_group`.

        Parameters
        ----------
        construction_state : :class:`.ConstructionState`
            The state of the current construction.

        edge_group : :class:`.EdgeGroup`
            The edge group for which a reaction should be found.

        Returns
        -------
        :class:`.Reaction`
            The reaction to use on the `edge_group`.

        """

        raise NotImplementedError()



import itertools
import logging
import typing
from collections.abc import Callable, Iterator, Sequence, Set

from stk._internal.ea.molecule_record import MoleculeRecord
from stk._internal.key_makers.molecule import MoleculeKeyMaker

from ..batch import Batch, BatchKey
from .yielded_batches import YieldedBatches

T = typing.TypeVar("T", bound=MoleculeRecord)
logger = logging.getLogger(__name__)

IncludedBatches: typing.TypeAlias = "Set[BatchKey] | None"
ExcludedBatches: typing.TypeAlias = "Set[BatchKey] | None"


class Selector(typing.Generic[T]):
    """
    An abstract base class for selectors.

    Selectors select batches of molecules from a population.
    Each batch is selected based on its fitness. The fitness of a
    batch is the sum of all fitness values of the molecules in the
    batch. Batches may be of size 1.

    Notes:

        You might notice that some of the public methods of this abstract
        base class are implemented. This is purely for convenience when
        implementing subclasses. The implemented public methods are simply
        default implementations, which can be safely ignored or overridden,
        when implementing subclasses. Any private methods are
        implementation details of these default implementations.

        *The Default Implementation*

        This section is only of use to people who want to add a new
        :class:`.Selector` subclass, and want to make use of the default
        implementation to make this job easier.

        When using the default implementation you do not need to
        implement :meth:`.select`, which is already provided, but instead
        :meth:`._select_from_batches` needs to be implemented. What the
        default implementation provides, is code, which does the batching
        of a `population` for you, which means you only have to worry
        about implementing the selection algorithm, which works on batches
        directly.

        The default implementation also automatically updates a
        :class:`.YieldedBatches` object for you, so that you can keep track
        of which batches have already been yielded, in case you want to
        prevent duplicate selection of batches or molecule records. Though
        whether you want to make use of this will depend on the nature of
        your selection algorithm.

    See Also:

        * :class:`.Batch`: Represents batches of selected molecules.

    Examples:

        *Subclass Implementation*

        The source code of the classes listed in :mod:`.selector` can serve
        as good examples.
    """

    def __init__(
        self,
        key_maker: MoleculeKeyMaker,
        fitness_modifier: Callable[[dict[T, float]], dict[T, float]],
        batch_size: int,
    ) -> None:
        """
        Parameters:
            key_maker:
                Used to get the keys of molecules, which are used to
                determine if two molecule records are duplicates of each
                other.

            fitness_modifier:
                Takes the `population` on which :meth:`.select` is called
                and returns a :class:`dict`, which maps records in the
                `population` to the fitness values the :class:`.Selector`
                should use.

            batch_size:
                The number of molecules yielded at once.
        """
        self._key_maker = key_maker
        self._fitness_modifier = fitness_modifier
        self._batch_size = batch_size

    def select(
        self,
        population: dict[T, float],
        included_batches: "IncludedBatches" = None,
        excluded_batches: "ExcludedBatches" = None,
    ) -> Iterator[Batch[T]]:
        """
        Yield batches of molecule records from `population`.

        Parameters:
            population:
                A collection of molecules from which batches are selected.

            included_batches:
                The identity keys of batches which are allowed to be
                yielded, if ``None`` all batches can be yielded. If not
                ``None`` only batches `included_batches` will be yielded.

            excluded_batches:
                The identity keys of batches which are not allowed to be
                yielded. If ``None``, no batch is forbidden from being
                yielded.

        Yields:
            A batch of selected molecule records.
        """
        batches = tuple(
            self._get_batches(
                population=self._fitness_modifier(population),
                included_batches=included_batches,
                excluded_batches=excluded_batches,
            )
        )

        yielded_batches: YieldedBatches[T] = YieldedBatches(self._key_maker)
        for batch in self._select_from_batches(
            batches=batches,
            yielded_batches=yielded_batches,
        ):
            yielded_batches.update(batch)
            yield batch

        cls_name = self.__class__.__name__
        logger.debug(
            f"{cls_name} yielded {yielded_batches.get_num()} batches."
        )

    def _get_batches(
        self,
        population: dict[T, float],
        included_batches: Set[BatchKey] | None,
        excluded_batches: Set[BatchKey] | None,
    ) -> Iterator[Batch[T]]:
        """
        Get batches molecules from `population`.

        Parameters:

            population:
                The molecule records which are to be batched.

            included_batches:
                The identity keys of batches which are allowed to be
                yielded, if ``None`` all batches can be yielded. If not
                ``None`` only batches `included_batches` will be yielded.

            excluded_batches:
                The identity keys of batches which are not allowed to be
                yielded. If ``None``, no batch is forbidden from being
                yielded.

            Yields:
                A batch of molecules from `population`.
        """

        def is_included(batch: Batch[T]) -> bool:
            if included_batches is None:
                return True
            return batch.get_identity_key() in included_batches

        def is_excluded(batch: Batch[T]) -> bool:
            if excluded_batches is None:
                return False
            return batch.get_identity_key() in excluded_batches

        for records in itertools.combinations(population, self._batch_size):
            batch = Batch(
                records=((record, population[record]) for record in records),
                key_maker=self._key_maker,
            )
            if is_included(batch) and not is_excluded(batch):
                yield batch

    def _select_from_batches(
        self,
        batches: Sequence[Batch[T]],
        yielded_batches: YieldedBatches[T],
    ) -> Iterator[Batch[T]]:
        """
        Yield batches from `batches`.

        Parameters:
            batches:
                Batches from which some are selected.
            yielded_batches:
                Keeps track of which batches have been yielded. This
                object automatically updates each time ``yield`` is called.
        Yields:
            A selected batch.
        """
        raise NotImplementedError()



def test_get_id(case_data):
    """
    Test :meth:`.Atom.get_id`.

    Parameters
    ----------
    case_data : :class:`case_data`
        A test case. Holds the atom to test and its correct id.

    Returns
    -------
    None : :class:`NoneType`

    """

    assert case_data.atom.get_id() == case_data.id



import pytest
import stk

from ..case_data import CaseData


class LinearData:
    """
    Data to initialize a linear polymer :class:`.CaseData` instance.

    Attributes
    ----------
    building_blocks : :class:`tuple` of :class:`.BuildingBlock`
        The building blocks of the polymer.

    repeating_unit : :class:`str` or :class:`tuple` of :class:`int`
        Passed to the `repeating_unit` parameter of
        :meth:`.polymer.Linear.__init__`.

    num_repeating_units : :class:`int`
        Passed to the `num_repeating_units` parameter of
        :meth:`.polymer.Linear.__init__`.

    num_new_atoms : :class:`int`
        The number of new atoms added by the construction process.

    num_new_bonds : :class:`int`
        The number of new bonds added by the construction process.

    num_building_blocks : :class:`dict`
        For each building block in :attr:`building_blocks`, maps its
        index to the number of times its used in the construction of
        the polymer.

    """

    def __init__(
        self,
        building_blocks,
        repeating_unit,
        num_repeating_units,
        num_new_atoms,
        num_new_bonds,
        num_building_blocks,
    ):
        """
        Initialize a :class:`.LinearData` instance.

        Parameters
        ----------
        building_blocks : :class:`tuple` of :class:`.BuildingBlock`
            The building blocks of the polymer.

        repeating_unit : :class:`str` or :class:`tuple` of :class:`int`
            Passed to the `repeating_unit` parameter of
            :meth:`.polymer.Linear.__init__`.

        num_repeating_units : :class:`int`
            Passed to the `num_repeating_units` parameter of
            :meth:`.polymer.Linear.__init__`.

        num_new_atoms : :class:`int`
            The number of new atoms added by the construction process.

        num_new_bonds : :class:`int`
            The number of new bonds added by the construction process.

        num_building_blocks : :class:`dict`
            For each building block in `building_blocks`, maps its
            index to the number of times its used in the construction
            of the polymer.

        """

        self.constructed_molecule = stk.ConstructedMolecule(
            topology_graph=stk.polymer.Linear(
                building_blocks=building_blocks,
                repeating_unit=repeating_unit,
                num_repeating_units=num_repeating_units,
            ),
        )
        self.num_new_atoms = num_new_atoms
        self.num_new_bonds = num_new_bonds
        self.num_building_blocks = {
            building_blocks[index]: num
            for index, num in num_building_blocks.items()
        }
        self.building_blocks = building_blocks


@pytest.fixture(
    scope="session",
    params=(
        lambda: LinearData(
            building_blocks=(
                stk.BuildingBlock("BrC#CBr", [stk.BromoFactory()]),
                stk.BuildingBlock(
                    smiles="Br[C+]=NC#CBr",
                    functional_groups=[stk.BromoFactory()],
                ),
            ),
            repeating_unit="AB",
            num_repeating_units=2,
            num_new_atoms=0,
            num_new_bonds=3,
            num_building_blocks={0: 2, 1: 2},
        ),
    ),
)
def linear_data(request) -> LinearData:
    """
    A :class:`.LinearData` instance.

    """

    return request.param()


@pytest.fixture
def linear(linear_data):
    """
    A :class:`.CaseData` instance.

    """

    return CaseData(
        constructed_molecule=linear_data.constructed_molecule,
        num_new_atoms=linear_data.num_new_atoms,
        num_new_bonds=linear_data.num_new_bonds,
        num_building_blocks=linear_data.num_building_blocks,
        building_blocks=linear_data.building_blocks,
    )



from typing import Any

import pytest
import stk

from .case_data import CaseData


def get_topology_graph(num_repeating_units: int) -> stk.TopologyGraph:
    return stk.polymer.Linear(
        building_blocks=[
            stk.BuildingBlock("BrCCBr", stk.BromoFactory()),
        ],
        repeating_unit="A",
        num_repeating_units=num_repeating_units,
    )


@pytest.fixture(
    scope="session",
    params=(
        lambda: CaseData(  # type: ignore
            selector=stk.Roulette(num_batches=50),
            population={
                stk.MoleculeRecord(
                    topology_graph=get_topology_graph(2),
                ): 1,
                stk.MoleculeRecord(
                    topology_graph=get_topology_graph(3),
                ): 2,
                stk.MoleculeRecord(
                    topology_graph=get_topology_graph(4),
                ): 3,
                stk.MoleculeRecord(
                    topology_graph=get_topology_graph(5),
                ): 4,
                stk.MoleculeRecord(
                    topology_graph=get_topology_graph(6),
                ): 5,
                stk.MoleculeRecord(
                    topology_graph=get_topology_graph(7),
                ): 6,
                stk.MoleculeRecord(
                    topology_graph=get_topology_graph(8),
                ): 7,
                stk.MoleculeRecord(
                    topology_graph=get_topology_graph(9),
                ): 8,
                stk.MoleculeRecord(
                    topology_graph=get_topology_graph(10),
                ): 9,
                stk.MoleculeRecord(
                    topology_graph=get_topology_graph(11),
                ): 10,
                stk.MoleculeRecord(
                    topology_graph=get_topology_graph(12),
                ): 11,
                stk.MoleculeRecord(
                    topology_graph=get_topology_graph(13),
                ): 12,
                stk.MoleculeRecord(
                    topology_graph=get_topology_graph(14),
                ): 13,
                stk.MoleculeRecord(
                    topology_graph=get_topology_graph(15),
                ): 14,
                stk.MoleculeRecord(
                    topology_graph=get_topology_graph(16),
                ): 15,
                stk.MoleculeRecord(
                    topology_graph=get_topology_graph(17),
                ): 16,
                stk.MoleculeRecord(
                    topology_graph=get_topology_graph(18),
                ): 17,
                stk.MoleculeRecord(
                    topology_graph=get_topology_graph(19),
                ): 18,
                stk.MoleculeRecord(
                    topology_graph=get_topology_graph(20),
                ): 19,
                stk.MoleculeRecord(
                    topology_graph=get_topology_graph(21),
                ): 20,
            },
        ),
    ),
)
def case_data(
    request: pytest.FixtureRequest,
) -> CaseData[stk.MoleculeRecord[Any]]:
    return request.param()



import numpy as np
import stk

from ...utilities import (
    get_displacement_vector,
    has_same_structure,
    is_clone,
)


def test_with_rotation_to_minimize_angle(molecule):
    """
    Test :meth:`.Molecule.with_rotation_to_minimize_angle`.

    Parameters
    ----------
    molecule : :class:`.Molecule`
        The molecule to test.

    Returns
    -------
    None : :class:`NoneType`

    """

    # Use to check that immutability is not violated.
    clone = molecule.clone()
    _test_with_rotation_to_minimize_angle(molecule)
    has_same_structure(molecule, clone)


def _test_with_rotation_to_minimize_angle(molecule):
    """
    Test :meth:`.Molecule.with_rotation_to_minimize_angle`.

    Parameters
    ----------
    molecule : :class:`.Molecule`
        The molecule to test.

    Returns
    -------
    None : :class:`NoneType`

    """

    start = get_displacement_vector(molecule, 0, 1)
    target = get_displacement_vector(molecule, 0, 2)
    new = molecule.with_rotation_to_minimize_angle(
        start=start,
        target=target,
        axis=stk.normalize_vector(np.cross(start, target)),
        origin=next(molecule.get_atomic_positions((0,))),
    )
    is_clone(new, molecule)

    result = get_displacement_vector(new, 0, 1)
    assert np.allclose(
        a=stk.normalize_vector(result),
        b=stk.normalize_vector(target),
        atol=1e-12,
    )
    assert abs(np.linalg.norm(start) - np.linalg.norm(result)) < 1e-15



import itertools as it

import pytest
import stk

from ..case_data import CaseData


@pytest.fixture
def one_one_reaction(
    functional_group1,
    functional_group1_2,
    bond_order,
    periodicity,
):
    """
    A :class:`.CaseData` instance.

    """

    return CaseData(
        reaction=stk.OneOneReaction(
            functional_group1=functional_group1,
            functional_group2=functional_group1_2,
            bond_order=bond_order,
            periodicity=periodicity,
        ),
        new_atoms=(),
        new_bonds=(
            get_bond(
                functional_group1=functional_group1,
                functional_group2=functional_group1_2,
                bond_order=bond_order,
                periodicity=periodicity,
            ),
        ),
        deleted_atoms=tuple(
            it.chain(
                functional_group1.get_deleters(),
                functional_group1_2.get_deleters(),
            )
        ),
        deleted_bonds=(),
    )


def get_bond(
    functional_group1,
    functional_group2,
    bond_order,
    periodicity,
):
    return stk.Bond(
        atom1=next(functional_group1.get_bonders()),
        atom2=next(functional_group2.get_bonders()),
        order=bond_order,
        periodicity=periodicity,
    )



"""
Bromo Factory
=============

"""


from stk._internal.functional_group_factories.functional_group_factory import (
    FunctionalGroupFactory,
)
from stk._internal.functional_group_factories.utilities import get_atom_ids
from stk._internal.functional_groups.bromo import Bromo


class BromoFactory(FunctionalGroupFactory):
    """
    Creates :class:`.Bromo` instances.

    Creates functional groups from substructures, which match the
    ``[*][Br]`` functional group string.

    Examples
    --------
    *Creating Functional Groups with the Factory*

    You want to create a building block which has :class:`.Bromo`
    functional groups. You want the atom bonded to the bromine to be
    the *bonder* atom, and the bromine atom to be the *deleter* atom.

    .. testcode:: creating-functional-groups-with-the-factory

        import stk

        building_block = stk.BuildingBlock(
            smiles='BrCCCBr',
            functional_groups=(stk.BromoFactory(), ),
        )

    .. testcode:: creating-functional-groups-with-the-factory
        :hide:

        assert all(
            isinstance(functional_group, stk.Bromo)
            for functional_group
            in building_block.get_functional_groups()
        )
        assert building_block.get_num_functional_groups() == 2

    See Also
    --------
    :class:`.GenericFunctionalGroup`
        Defines *bonders* and  *deleters*.

    """

    def __init__(self, bonders=(0,), deleters=(1,), placers=None):
        """
        Initialize a :class:`.BromoFactory` instance.

        Parameters
        ----------
        bonders : :class:`tuple` of :class:`int`
            The indices of atoms in the functional group string, which
            are *bonder* atoms.

        deleters : :class:`tuple` of :class:`int`
            The indices of atoms in the functional group string, which
            are *deleter* atoms.

        placers : :class:`tuple` of :class:`int`, optional
            The indices of atoms in the functional group string, which
            are *placer* atoms. If ``None``, `bonders` will be used.

        """

        self._bonders = bonders
        self._deleters = deleters
        self._placers = bonders if placers is None else placers

    def get_functional_groups(self, molecule):
        for atom_ids in get_atom_ids("[*][Br]", molecule):
            atoms = tuple(molecule.get_atoms(atom_ids))
            yield Bromo(
                bromine=atoms[1],
                atom=atoms[0],
                bonders=tuple(atoms[i] for i in self._bonders),
                deleters=tuple(atoms[i] for i in self._deleters),
                placers=tuple(atoms[i] for i in self._placers),
            )



import pathos

from stk._internal.construction_state.construction_state import (
    ConstructionState,
)

from .utilities import _Placement


class _Parallel:
    """
    Holds parallel implementations of topology graph methods.
    """

    def __init__(
        self, stages: tuple[tuple[int, ...], ...], num_processes: int
    ) -> None:
        """
        Initialize a :class:`._Parallel` instance.

        Parameters:
            stages:
                A :class:`tuple` of the form ``((v1, v2, v3), (v4, v5))``,
                where each nested :class:`tuple` holds the
                :class:`.Vertex` objects used for placement in a particular
                stage.

            num_processes:
                The number of parallel processes to spawn.
        """
        self._stages = stages
        self._num_processes = num_processes

    def _place_building_blocks(
        self,
        state: ConstructionState,
    ) -> ConstructionState:
        with pathos.pools.ProcessPool(self._num_processes) as pool:
            for stage in self._stages:
                vertices = tuple(state.get_vertices(stage))
                building_blocks = tuple(map(state.get_building_block, stage))
                edges = tuple(map(state.get_edges, stage))
                placements = map(
                    _Placement,
                    vertices,
                    edges,
                    building_blocks,
                )
                placement_results = pool.map(
                    lambda placement: placement.get_result(),
                    placements,
                )
                state = state.with_placement_results(
                    vertices=vertices,
                    edges=edges,
                    building_blocks=building_blocks,
                    results=placement_results,
                )
        return state

    def get_num_stages(self) -> int:
        """
        Get the number of placement stages.

        Returns:
            The number of placement stages.
        """
        return len(self._stages)



"""
Four Plus Four
==============

"""

from stk._internal.topology_graphs.edge import Edge

from .cage import Cage
from .vertices import NonLinearVertex


class FourPlusFour(Cage):
    """
    Represents a cube cage topology graph.

    Unoptimized construction

    .. moldoc::

        import moldoc.molecule as molecule
        import stk

        bb = stk.BuildingBlock(
            smiles='Brc1cc(Br)cc(Br)c1',
            functional_groups=[stk.BromoFactory()],
        )
        cage = stk.ConstructedMolecule(
            topology_graph=stk.cage.FourPlusFour((bb, )),
        )

        moldoc_display_molecule = molecule.Molecule(
            atoms=(
                molecule.Atom(
                    atomic_number=atom.get_atomic_number(),
                    position=position,
                ) for atom, position in zip(
                    cage.get_atoms(),
                    cage.get_position_matrix(),
                )
            ),
            bonds=(
                molecule.Bond(
                    atom1_id=bond.get_atom1().get_id(),
                    atom2_id=bond.get_atom2().get_id(),
                    order=bond.get_order(),
                ) for bond in cage.get_bonds()
            ),
        )

    :class:`.MCHammer` optimized construction

    .. moldoc::

        import moldoc.molecule as molecule
        import stk

        bb = stk.BuildingBlock(
            smiles='Brc1cc(Br)cc(Br)c1',
            functional_groups=[stk.BromoFactory()],
        )
        cage = stk.ConstructedMolecule(
            topology_graph=stk.cage.FourPlusFour(
                building_blocks=(bb, ),
                optimizer=stk.MCHammer(),
            ),
        )

        moldoc_display_molecule = molecule.Molecule(
            atoms=(
                molecule.Atom(
                    atomic_number=atom.get_atomic_number(),
                    position=position,
                ) for atom, position in zip(
                    cage.get_atoms(),
                    cage.get_position_matrix(),
                )
            ),
            bonds=(
                molecule.Bond(
                    atom1_id=bond.get_atom1().get_id(),
                    atom2_id=bond.get_atom2().get_id(),
                    order=bond.get_order(),
                ) for bond in cage.get_bonds()
            ),
        )

    Building blocks with three functional groups are required for
    this topology.

    When using a :class:`dict` for the `building_blocks` parameter,
    as in :ref:`cage-topology-graph-examples`:
    *Multi-Building Block Cage Construction*, a
    :class:`.BuildingBlock`, with the following number of functional
    groups, needs to be assigned to each of the following vertex ids:

        | 3-functional groups: 0 to 7

    See :class:`.Cage` for more details and examples.

    """

    _x = 1
    _vertex_prototypes = (
        NonLinearVertex(0, [-_x, _x, -_x], False),
        NonLinearVertex(1, [-_x, -_x, -_x], False),
        NonLinearVertex(2, [_x, _x, -_x], False),
        NonLinearVertex(3, [_x, -_x, -_x], False),
        NonLinearVertex(4, [-_x, _x, _x], False),
        NonLinearVertex(5, [-_x, -_x, _x], False),
        NonLinearVertex(6, [_x, _x, _x], False),
        NonLinearVertex(7, [_x, -_x, _x], False),
    )

    _edge_prototypes = (
        Edge(0, _vertex_prototypes[0], _vertex_prototypes[1]),
        Edge(1, _vertex_prototypes[0], _vertex_prototypes[2]),
        Edge(2, _vertex_prototypes[0], _vertex_prototypes[4]),
        Edge(3, _vertex_prototypes[1], _vertex_prototypes[3]),
        Edge(4, _vertex_prototypes[1], _vertex_prototypes[5]),
        Edge(5, _vertex_prototypes[2], _vertex_prototypes[6]),
        Edge(6, _vertex_prototypes[2], _vertex_prototypes[3]),
        Edge(7, _vertex_prototypes[3], _vertex_prototypes[7]),
        Edge(8, _vertex_prototypes[4], _vertex_prototypes[6]),
        Edge(9, _vertex_prototypes[4], _vertex_prototypes[5]),
        Edge(10, _vertex_prototypes[5], _vertex_prototypes[7]),
        Edge(11, _vertex_prototypes[6], _vertex_prototypes[7]),
    )

    _num_windows = 6
    _num_window_types = 1



import pytest
import stk

from ....case_data import CaseData
from ...building_blocks import get_linker, get_pd_atom


@pytest.fixture(
    scope="session",
    params=(
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                topology_graph=stk.cage.M3L6(
                    building_blocks={
                        get_pd_atom(): range(3),
                        get_linker(): range(3, 9),
                    },
                    reaction_factory=stk.DativeReactionFactory(
                        reaction_factory=stk.GenericReactionFactory(
                            bond_orders={
                                frozenset(
                                    {
                                        stk.GenericFunctionalGroup,
                                        stk.SingleAtom,
                                    }
                                ): 9,
                            },
                        ),
                    ),
                ),
            ),
            smiles=(
                "[H]C1=C([H])C2=C([H])C(=C1[H])C1=C([H])C([H])=N(->["
                "Pd+2]34<-N5=C([H])C([H])=C(C([H])=C5[H])C5=C([H])C("
                "[H])=C([H])C(=C5[H])C5=C([H])C([H])=N(->[Pd+2]6(<-N"
                "7=C([H])C([H])=C(C([H])=C7[H])C7=C([H])C([H])=C([H]"
                ")C(=C7[H])C7=C([H])C([H])=N(->[Pd+2](<-N8=C([H])C(["
                "H])=C(C([H])=C8[H])C8=C([H])C(=C([H])C([H])=C8[H])C"
                "8=C([H])C([H])=N->3C([H])=C8[H])(<-N3=C([H])C([H])="
                "C(C([H])=C3[H])C3=C([H])C(=C([H])C([H])=C3[H])C3=C("
                "[H])C([H])=N->4C([H])=C3[H])<-N3=C([H])C([H])=C(C(["
                "H])=C3[H])C3=C([H])C(=C([H])C([H])=C3[H])C3=C([H])C"
                "([H])=N->6C([H])=C3[H])C([H])=C7[H])<-N3=C([H])C([H"
                "])=C2C([H])=C3[H])C([H])=C5[H])C([H])=C1[H]"
            ),
            name=name,
        ),
    ),
)
def metal_cage_m3l6(request) -> CaseData:
    return request.param(
        f"{request.fixturename}{request.param_index}",
    )



"""
Metal Complex Vertices
======================

"""

from scipy.spatial.distance import euclidean

from stk._internal.topology_graphs.vertex import Vertex
from stk._internal.utilities.utilities import get_projection


class UnaligningVertex(Vertex):
    """
    A vertex which does no rotations.

    """

    def place_building_block(self, building_block, edges):
        return building_block.with_centroid(
            position=self._position,
            atom_ids=building_block.get_placer_ids(),
        ).get_position_matrix()

    def map_functional_groups_to_edges(self, building_block, edges):
        return {fg_id: edge.get_id() for fg_id, edge in enumerate(edges)}


class MetalVertex(UnaligningVertex):
    """
    Places the metal in a :class:`.MetalComplex`.

    """

    pass


class MonoDentateLigandVertex(Vertex):
    """
    Places monodentate ligand in a :class:`.MetalComplex`.

    """

    def place_building_block(self, building_block, edges):
        building_block = building_block.with_centroid(
            position=self._position,
            atom_ids=building_block.get_placer_ids(),
        )
        assert building_block.get_num_functional_groups() == 1, (
            f"{building_block} needs to have exactly 1 functional "
            "group but has "
            f"{building_block.get_num_functional_groups()}."
        )
        (fg,) = building_block.get_functional_groups(0)
        fg_centroid = building_block.get_centroid(
            atom_ids=fg.get_placer_ids(),
        )
        core_centroid = building_block.get_centroid(
            atom_ids=building_block.get_core_atom_ids(),
        )
        edge_centroid = sum(edge.get_position() for edge in edges) / len(edges)
        return building_block.with_rotation_between_vectors(
            start=fg_centroid - core_centroid,
            target=edge_centroid - self._position,
            origin=self._position,
        ).get_position_matrix()

    def map_functional_groups_to_edges(self, building_block, edges):
        return {0: edges[0].get_id()}


class BiDentateLigandVertex(Vertex):
    """
    Places bidentate ligand in a :class:`.MetalComplex`.

    """

    def place_building_block(self, building_block, edges):
        building_block = building_block.with_centroid(
            position=self._position,
            atom_ids=building_block.get_placer_ids(),
        )
        assert building_block.get_num_functional_groups() == 2, (
            f"{building_block} needs to have exactly 2 functional "
            "groups but has "
            f"{building_block.get_num_functional_groups()}."
        )

        fg0_position, fg1_position = (
            building_block.get_centroid(fg.get_placer_ids())
            for fg in building_block.get_functional_groups()
        )
        edge_position1, edge_position2 = (
            edge.get_position() for edge in edges
        )
        building_block = building_block.with_rotation_between_vectors(
            start=fg1_position - fg0_position,
            target=edge_position2 - edge_position1,
            origin=building_block.get_centroid(),
        )

        placer_centroid = building_block.get_centroid(
            atom_ids=building_block.get_placer_ids(),
        )
        core_centroid = building_block.get_centroid(
            atom_ids=building_block.get_core_atom_ids(),
        )
        core_to_placer = placer_centroid - core_centroid

        fg0_position, fg1_position = (
            building_block.get_centroid(fg.get_placer_ids())
            for fg in building_block.get_functional_groups()
        )
        fg_vector = fg1_position - fg0_position

        fg_vector_projection = get_projection(
            start=core_to_placer,
            target=fg_vector,
        )

        edge_centroid = sum(edge.get_position() for edge in edges) / len(edges)
        building_block = building_block.with_rotation_between_vectors(
            start=core_to_placer - fg_vector_projection,
            target=edge_centroid - self._position,
            origin=building_block.get_centroid(),
        )

        return building_block.with_centroid(
            position=self._position,
            atom_ids=building_block.get_placer_ids(),
        ).get_position_matrix()

    def map_functional_groups_to_edges(self, building_block, edges):
        (fg,) = building_block.get_functional_groups(0)
        fg_position = building_block.get_centroid(fg.get_placer_ids())

        def fg_distance(edge):
            return euclidean(edge.get_position(), fg_position)

        edges = sorted(edges, key=fg_distance)
        return {fg_id: edge.get_id() for fg_id, edge in enumerate(edges)}



import pytest
import stk

from ...case_data import CaseData


@pytest.fixture(
    scope="session",
    params=(
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                topology_graph=stk.cof.PeriodicSquare(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles="BrC1=C(Br)[C+]=N1",
                            functional_groups=[stk.BromoFactory()],
                        ),
                        stk.BuildingBlock(
                            smiles="BrC1=C(Br)C(F)(Br)[C+]1Br",
                            functional_groups=[stk.BromoFactory()],
                        ),
                    ),
                    lattice_size=(2, 2, 1),
                ),
            ),
            smiles=(
                "FC12C3=C(N=[C+]3)C3=C4C5=C([C+]=N5)[C+]5C6=C7C8=C([C+"
                "]=N8)[C+]3C4(F)C3=C(N=[C+]3)C3=C1C1=C([C+]=N1)[C+]1C("
                "=C(C4=C([C+]=N4)[C+]32)C1(F)C1=C6N=[C+]1)C1=C([C+]=N1"
                ")C75F"
            ),
            name=name,
        ),
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                topology_graph=stk.cof.PeriodicSquare(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles="BrC1=C(Br)[C+]=N1",
                            functional_groups=[stk.BromoFactory()],
                        ),
                        stk.BuildingBlock(
                            smiles="BrC1=C(Br)C(F)(Br)[C+]1Br",
                            functional_groups=[stk.BromoFactory()],
                        ),
                    ),
                    lattice_size=(2, 2, 1),
                    optimizer=stk.PeriodicCollapser(),
                ),
            ),
            smiles=(
                "FC12C3=C(N=[C+]3)C3=C4C5=C([C+]=N5)[C+]5C6=C7C8=C([C+"
                "]=N8)[C+]3C4(F)C3=C(N=[C+]3)C3=C1C1=C([C+]=N1)[C+]1C("
                "=C(C4=C([C+]=N4)[C+]32)C1(F)C1=C6N=[C+]1)C1=C([C+]=N1"
                ")C75F"
            ),
            name=name,
        ),
    ),
)
def cof_periodic_square(request) -> CaseData:
    return request.param(
        f"{request.fixturename}{request.param_index}",
    )



from stk._internal.atom import Atom
from stk._internal.molecule import Molecule


class AtomInfo:
    """
    Holds additional info about :class:`.ConstructedMolecule` atoms.
    """

    def __init__(
        self,
        atom: Atom,
        building_block_atom: Atom | None,
        building_block: Molecule | None,
        building_block_id: int | None,
    ) -> None:
        """
        Parameters:

            atom:
                The atom about which information is held.

            building_block_atom:
                The building block atom from which this atom
                originates. Can be ``None``, if the atom was not part
                of the building block, but was added by the
                construction process instead.

            building_block:
                The building block from which the atom originates.
                Can be ``None``, if the atom was not part of a building
                block, but was added by the construction process
                instead.

            building_block_id:
                A unique id for each :class:`.Molecule` placed during
                the construction of the :class:`.ConstructedMolecule`.
                As a single :class:`.Molecule` can be placed multiple
                times during construction, the `building_block_id`
                allows the user to distinguish between each placement.
                Can be ``None``, if the atom was not part of a
                building block, but was added by the construction
                process instead.
        """
        self._atom = atom
        self._building_block_atom = building_block_atom
        self._building_block = building_block
        self._building_block_id = building_block_id

    def get_atom(self) -> Atom:
        """
        Get the atom about which information is held.

        Returns:

            The atom.

        """

        return self._atom

    def get_building_block_atom(self) -> Atom | None:
        """
        Get the original atom held by the building block.

        Returns:

            The building block atom or ``None`` if the atom was not
            originally found in a building block, but was added by the
            construction process instead.

        """

        return self._building_block_atom

    def get_building_block(self) -> Molecule | None:
        """
        Get the building block from which the atom originates.

        Returns:

            The building block or ``None`` if the atom was not
            originally found in a building block, but was added by the
            construction process instead.

        """

        return self._building_block

    def get_building_block_id(self) -> int | None:
        """
        Get the id of the atom's building block.

        A unique id for each :class:`.Molecule` placed during
        the construction of the :class:`.ConstructedMolecule`. As a
        single :class:`.Molecule` can be placed multiple times
        during construction, the building block id allows
        the user to distinguish between each placement.

        Returns:

            The unique building block id or ``None`` if the atom was
            not originally found in a building block, but was added by
            the construction process instead.

        """

        return self._building_block_id



import itertools as it


def are_equivalent_atoms(atoms1, atoms2):
    for atom1, atom2 in it.zip_longest(atoms1, atoms2):
        is_equivalent_atom(atom1, atom2)


def is_equivalent_atom(atom1, atom2):
    assert atom1.get_id() == atom2.get_id()
    assert atom1.__class__ is atom2.__class__
    assert atom1.get_atomic_number() == atom2.get_atomic_number()
    assert atom1.get_charge() == atom2.get_charge()


def are_equivalent_bonds(bonds1, bonds2):
    for bond1, bond2 in it.zip_longest(bonds1, bonds2):
        is_equivalent_bond(bond1, bond2)


def is_equivalent_bond(bond1, bond2):
    assert bond1.__class__ is bond2.__class__
    assert bond1.get_order() == bond2.get_order()
    assert bond1.get_periodicity() == bond2.get_periodicity()
    is_equivalent_atom(bond1.get_atom1(), bond2.get_atom1())
    is_equivalent_atom(bond1.get_atom2(), bond2.get_atom2())



from .cof import *  # noqa
from .hexagonal import *  # noqa
from .honeycomb import *  # noqa
from .kagome import *  # noqa
from .linkerless_honeycomb import *  # noqa
from .periodic_hexagonal import *  # noqa
from .periodic_honeycomb import *  # noqa
from .periodic_kagome import *  # noqa
from .periodic_linkerless_honeycomb import *  # noqa
from .periodic_square import *  # noqa
from .square import *  # noqa



"""
Difluoro Factory
================

"""


from stk._internal.functional_group_factories.functional_group_factory import (
    FunctionalGroupFactory,
)
from stk._internal.functional_group_factories.utilities import get_atom_ids
from stk._internal.functional_groups.difluoro import Difluoro


class DifluoroFactory(FunctionalGroupFactory):
    """
    Creates :class:`.Difluoro` instances.

    Creates functional groups from substructures, which match the
    ``[F][#6]~[#6][F]`` functional group string.

    Examples
    --------
    *Creating Functional Groups with the Factory*

    You want to create a building block which has :class:`.Difluoro`
    functional groups. You want the non-fluorine atoms in those
    functional groups to be the *bonder* atoms, and the fluorine atoms
    to be the *deleter* atoms.

    .. testcode:: creating-functional-groups-with-the-factory

        import stk

        building_block = stk.BuildingBlock(
            smiles='FCC(F)CCC',
            functional_groups=(stk.DifluoroFactory(), ),
        )

    .. testcode:: creating-functional-groups-with-the-factory
        :hide:

        assert all(
            isinstance(functional_group, stk.Difluoro)
            for functional_group
            in building_block.get_functional_groups()
        )
        assert building_block.get_num_functional_groups() == 1

    *Changing the Bonder and Deleter Atoms*

    You want to create a building block which has :class:`.Difluoro`
    functional groups, You want only one of non-fluorine atoms to be
    a *bonder* atom and its neighboring fluorine atom to be a
    *deleter* atom.

    .. testcode:: changing-the-bonder-and-deleter-atoms

        import stk

        difluoro_factory = stk.DifluoroFactory(
            # The index of one of the non-fluorine atoms in the
            # functional group string (see docstring) is 1.
            bonders=(1, ),
            # The neighboring fluorine atom has an index of 0.
            deleters=(0, ),
        )
        building_block = stk.BuildingBlock(
            smiles='FCC(F)CCC',
            functional_groups=(difluoro_factory, ),
        )

    .. testcode:: changing-the-bonder-and-deleter-atoms
        :hide:

        fg, = building_block.get_functional_groups()
        assert fg.get_num_bonders() == 1
        assert sum(1 for _ in fg.get_deleters()) == 1

        assert all(
            isinstance(atom, stk.C)
            for functional_group
            in building_block.get_functional_groups()
            for atom
            in functional_group.get_bonders()
        )
        assert all(
            isinstance(atom, stk.F)
            for functional_group
            in building_block.get_functional_groups()
            for atom
            in functional_group.get_deleters()
        )

    See Also
    --------
    :class:`.GenericFunctionalGroup`
        Defines *bonders* and  *deleters*.

    """

    def __init__(self, bonders=(1, 2), deleters=(0, 3), placers=None):
        """
        Initialize a :class:`.DifluoroFactory` instance.

        Parameters
        ----------
        bonders : :class:`tuple` of :class:`int`
            The indices of atoms in the functional group string, which
            are *bonder* atoms.

        deleters : :class:`tuple` of :class:`int`
            The indices of atoms in the functional group string, which
            are *deleter* atoms.

        placers : :class:`tuple` of :class:`int`, optional
            The indices of atoms in the functional group string, which
            are *placer* atoms. If ``None``, `bonders` will be used.

        """

        self._bonders = bonders
        self._deleters = deleters
        self._placers = bonders if placers is None else placers

    def get_functional_groups(self, molecule):
        for atom_ids in get_atom_ids("[F][#6]~[#6][F]", molecule):
            atoms = tuple(molecule.get_atoms(atom_ids))
            yield Difluoro(
                atom1=atoms[1],
                fluorine1=atoms[0],
                atom2=atoms[2],
                fluorine2=atoms[3],
                bonders=tuple(atoms[i] for i in self._bonders),
                deleters=tuple(atoms[i] for i in self._deleters),
                placers=tuple(atoms[i] for i in self._placers),
            )



def test_get_num_functional_groups(case_data):
    """
    Test :meth:`.BuildingBlock.get_num_functional_groups`.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        A test case. Holds the building block to test and the
        correct number of functional groups.

    Returns
    -------
    None : :class:`NoneType`

    """

    assert case_data.building_block.get_num_functional_groups() == len(
        case_data.functional_groups
    )



"""
Turbomole Updating Utilities
============================

"""

import numpy as np

from stk._internal.utilities.utilities import periodic_table


class _CoordSection:
    """
    Holds the coord section of a turbomole file.

    """

    def __init__(self, lines):
        """
        Initialize a :class:`._CoordSection`.

        This initializer assumes that coordinates used in `lines` are
        defined in angstroms.

        Parameters
        ----------
        lines : :class:`iterable` of :class:`str`
            The lines of the coord section.

        """

        elements = []
        position_matrix = []
        for line in lines:
            x, y, z, element = line.split()

            if element.isnumeric():
                element = periodic_table[int(element)]

            elements.append(element)
            position_matrix.append([float(x), float(y), float(z)])

        self._position_matrix = np.array(position_matrix)
        self._elements = tuple(elements)

    @classmethod
    def init_bohr(cls, lines):
        """
        Initialize a :class:`._CoordSection`.

        This initializer assumes that coordinates used in `lines`
        are defined in bohr units.

        Parameters
        ----------
        lines : :class:`iterable` of :class:`str`
            The lines of the coord section.

        Returns
        -------
        :class:`._CoordSection`
            The coord section.

        """

        bohr_to_ang = 0.5291772105638411

        obj = cls.__new__(cls)
        elements = []
        position_matrix = []
        for line in lines:
            x, y, z, element = line.split()

            if element.isnumeric():
                element = periodic_table[int(element)]

            elements.append(element)
            position_matrix.append(
                [
                    float(x) * bohr_to_ang,
                    float(y) * bohr_to_ang,
                    float(z) * bohr_to_ang,
                ]
            )

        obj._elements = tuple(elements)
        obj._position_matrix = np.array(position_matrix)
        return obj

    def get_position_matrix(self):
        """
        Get the position matrix defined in the coord section.

        Returns
        -------
        :class:`numpy.ndarray`
            The position matrix defined in the coord section.

        """

        return np.array(self._position_matrix)

    def get_num_atoms(self):
        """
        Get the number of atoms defined in the coord section.

        Returns
        -------
        :class:`int`
            The number of atoms defined in the coord section.

        """

        return len(self._elements)

    def get_elements(self):
        """
        Yield the elements of the atoms in the coord section.

        The elements are yielded in order of atom id.

        Yields
        ------
        :class:`str`
            The chemical symbol of an atom.

        """

        yield from self._elements


def _get_coord_section(path, num_atoms):
    """
    Get the coord section defined in `path`.

    Parameters
    ----------
    path : :class:`str`
        The path to the turbomole file.

    num_atoms : :class:`int`
        The number of atoms in the molecule.

    Returns
    -------
    :class:`._CoordSection`
        The coord section defined in `path`.

    Raises
    ------
    :class:`RuntimeError`
        If `path` uses fractional coordinates, as they are not
        currently supported.

    :class:`RuntimeError`
        If no coord section in found in `path`.

    """

    with open(path, "r") as f:
        content = f.readlines()

    for line_number, line in enumerate(content):
        if "$coord" in line:
            lines = content[line_number + 1 : line_number + 1 + num_atoms]

            if "angs" in line:
                return _CoordSection(lines)

            elif "frac" in line:
                raise RuntimeError(
                    f"{path} uses fractional coordinates, which are "
                    "not currently supported."
                )

            return _CoordSection.init_bohr(lines)

    raise RuntimeError(f"No coord section found in {path}.")


def _with_structure_from_turbomole(self, path):
    """
    Update the structure with one taken from a Turbomole file.

    Note that coordinates in ``.coord`` files can be given in Bohr or
    Angstrom, which is handled. Fractional coordinates are not
    currently handled.

    Parameters
    ----------
    path : :class:`str`
        The full path of the ``.coord`` file from which the
        structure should be updated.

    Returns
    -------
    :class:`.Molecule`
        The molecule is returned.

    Raises
    ------
    :class:`RuntimeError`
        If the number of atoms in the file does not match the
        number of atoms in the molecule.

    :class:`RuntimeError`
        If atom elements in the file do not agree with the atom
        elements in the molecule.

    """

    num_atoms = len(self._atoms)
    section = _get_coord_section(path, num_atoms)

    if section.get_num_atoms() != num_atoms:
        raise RuntimeError(
            f"The number of atoms in {path}, "
            f"{section.get_num_atoms()}, does not match the number "
            f"of atoms in the molecule, {num_atoms}."
        )

    for atom_id, (element, atom) in enumerate(
        zip(
            section.get_elements(),
            self._atoms,
        )
    ):
        if element != atom.__class__.__name__:
            raise RuntimeError(
                f"The element of atom {atom_id} in {path}, "
                f"{element}, does not match the element in the "
                f"molecule, {atom.__class__.__name__}."
            )

    return self._with_position_matrix(section.get_position_matrix())



import pytest
import stk

from ..case_data import CaseData


@pytest.fixture(
    scope="session",
    params=(
        lambda: CaseData(
            factory=stk.RingAmineFactory(),
            molecule=stk.BuildingBlock("NCC(Br)c1c(Br)cccc1"),
            functional_groups=(
                stk.RingAmine(
                    nitrogen=stk.N(0),
                    carbon1=stk.C(1),
                    carbon2=stk.C(2),
                    carbon3=stk.C(4),
                    hydrogen1=stk.H(11),
                    hydrogen2=stk.H(12),
                    hydrogen3=stk.H(15),
                ),
            ),
        ),
    ),
)
def case_data(request) -> CaseData:
    """
    A :class:`.CaseData` instance.

    """

    return request.param()
